# -*- coding: utf-8 -*-

from __future__ import absolute_import, print_function, unicode_literals

import hashlib
import logging
import os
import shutil
import signal
import tempfile
import textwrap
import time
import uuid

import psutil
import pytest
import salt.ext.six as six
import salt.utils.files
import salt.utils.path
import salt.utils.platform
import salt.utils.stringutils
from saltfactories.utils.ports import get_unused_localhost_port
from tests.support.case import ModuleCase
from tests.support.helpers import skip_if_not_root, slowTest, with_tempfile
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import skipIf

log = logging.getLogger(__name__)


@pytest.mark.windows_whitelisted
class CPModuleTest(ModuleCase):
    """
    Validate the cp module
    """

    def run_function(self, *args, **kwargs):  # pylint: disable=arguments-differ
        """
        Ensure that results are decoded

        TODO: maybe move this behavior to ModuleCase itself?
        """
        return salt.utils.data.decode(
            super(CPModuleTest, self).run_function(*args, **kwargs)
        )

    @with_tempfile()
    @slowTest
    def test_get_file(self, tgt):
        """
        cp.get_file
        """
        self.run_function("cp.get_file", ["salt://grail/scene33", tgt])
        with salt.utils.files.fopen(tgt, "r") as scene:
            data = salt.utils.stringutils.to_unicode(scene.read())
        self.assertIn("KNIGHT:  They're nervous, sire.", data)
        self.assertNotIn("bacon", data)

    @slowTest
    def test_get_file_to_dir(self):
        """
        cp.get_file
        """
        tgt = os.path.join(RUNTIME_VARS.TMP, "")
        self.run_function("cp.get_file", ["salt://grail/scene33", tgt])
        with salt.utils.files.fopen(tgt + "scene33", "r") as scene:
            data = salt.utils.stringutils.to_unicode(scene.read())
        self.assertIn("KNIGHT:  They're nervous, sire.", data)
        self.assertNotIn("bacon", data)

    @with_tempfile()
    @skipIf(
        salt.utils.platform.is_windows() and six.PY3,
        "This test hangs on Windows on Py3",
    )
    def test_get_file_templated_paths(self, tgt):
        """
        cp.get_file
        """
        self.run_function(
            "cp.get_file",
            [
                "salt://{{grains.test_grain}}",
                tgt.replace("cheese", "{{grains.test_grain}}"),
            ],
            template="jinja",
        )
        with salt.utils.files.fopen(tgt, "r") as cheese:
            data = salt.utils.stringutils.to_unicode(cheese.read())
        self.assertIn("Gromit", data)
        self.assertNotIn("bacon", data)

    @with_tempfile()
    @slowTest
    def test_get_file_gzipped(self, tgt):
        """
        cp.get_file
        """
        src = os.path.join(RUNTIME_VARS.FILES, "file", "base", "file.big")
        with salt.utils.files.fopen(src, "rb") as fp_:
            hash_str = hashlib.md5(fp_.read()).hexdigest()

        self.run_function("cp.get_file", ["salt://file.big", tgt], gzip=5)
        with salt.utils.files.fopen(tgt, "rb") as scene:
            data = scene.read()
        self.assertEqual(hash_str, hashlib.md5(data).hexdigest())
        data = salt.utils.stringutils.to_unicode(data)
        self.assertIn("KNIGHT:  They're nervous, sire.", data)
        self.assertNotIn("bacon", data)

    @slowTest
    def test_get_file_makedirs(self):
        """
        cp.get_file
        """
        tgt = os.path.join(RUNTIME_VARS.TMP, "make", "dirs", "scene33")
        self.run_function("cp.get_file", ["salt://grail/scene33", tgt], makedirs=True)
        self.addCleanup(
            shutil.rmtree, os.path.join(RUNTIME_VARS.TMP, "make"), ignore_errors=True
        )
        with salt.utils.files.fopen(tgt, "r") as scene:
            data = salt.utils.stringutils.to_unicode(scene.read())
        self.assertIn("KNIGHT:  They're nervous, sire.", data)
        self.assertNotIn("bacon", data)

    @with_tempfile()
    @slowTest
    def test_get_template(self, tgt):
        """
        cp.get_template
        """
        self.run_function(
            "cp.get_template", ["salt://grail/scene33", tgt], spam="bacon"
        )
        with salt.utils.files.fopen(tgt, "r") as scene:
            data = salt.utils.stringutils.to_unicode(scene.read())
        self.assertIn("bacon", data)
        self.assertNotIn("spam", data)

    @slowTest
    def test_get_dir(self):
        """
        cp.get_dir
        """
        tgt = os.path.join(RUNTIME_VARS.TMP, "many")
        self.run_function("cp.get_dir", ["salt://grail", tgt])
        self.assertIn("grail", os.listdir(tgt))
        self.assertIn("36", os.listdir(os.path.join(tgt, "grail")))
        self.assertIn("empty", os.listdir(os.path.join(tgt, "grail")))
        self.assertIn("scene", os.listdir(os.path.join(tgt, "grail", "36")))

    @slowTest
    def test_get_dir_templated_paths(self):
        """
        cp.get_dir
        """
        tgt = os.path.join(RUNTIME_VARS.TMP, "many")
        self.run_function(
            "cp.get_dir",
            ["salt://{{grains.script}}", tgt.replace("many", "{{grains.alot}}")],
        )
        self.assertIn("grail", os.listdir(tgt))
        self.assertIn("36", os.listdir(os.path.join(tgt, "grail")))
        self.assertIn("empty", os.listdir(os.path.join(tgt, "grail")))
        self.assertIn("scene", os.listdir(os.path.join(tgt, "grail", "36")))

    # cp.get_url tests

    @with_tempfile()
    @slowTest
    def test_get_url(self, tgt):
        """
        cp.get_url with salt:// source given
        """
        self.run_function("cp.get_url", ["salt://grail/scene33", tgt])
        with salt.utils.files.fopen(tgt, "r") as scene:
            data = salt.utils.stringutils.to_unicode(scene.read())
        self.assertIn("KNIGHT:  They're nervous, sire.", data)
        self.assertNotIn("bacon", data)

    @slowTest
    def test_get_url_makedirs(self):
        """
        cp.get_url
        """
        tgt = os.path.join(RUNTIME_VARS.TMP, "make", "dirs", "scene33")
        self.run_function("cp.get_url", ["salt://grail/scene33", tgt], makedirs=True)
        self.addCleanup(
            shutil.rmtree, os.path.join(RUNTIME_VARS.TMP, "make"), ignore_errors=True
        )
        with salt.utils.files.fopen(tgt, "r") as scene:
            data = salt.utils.stringutils.to_unicode(scene.read())
        self.assertIn("KNIGHT:  They're nervous, sire.", data)
        self.assertNotIn("bacon", data)

    @slowTest
    def test_get_url_dest_empty(self):
        """
        cp.get_url with salt:// source given and destination omitted.
        """
        ret = self.run_function("cp.get_url", ["salt://grail/scene33"])
        with salt.utils.files.fopen(ret, "r") as scene:
            data = salt.utils.stringutils.to_unicode(scene.read())
        self.assertIn("KNIGHT:  They're nervous, sire.", data)
        self.assertNotIn("bacon", data)

    @slowTest
    def test_get_url_no_dest(self):
        """
        cp.get_url with salt:// source given and destination set as None
        """
        tgt = None
        ret = self.run_function("cp.get_url", ["salt://grail/scene33", tgt])
        self.assertIn("KNIGHT:  They're nervous, sire.", ret)

    @slowTest
    def test_get_url_nonexistent_source(self):
        """
        cp.get_url with nonexistent salt:// source given
        """
        tgt = None
        ret = self.run_function("cp.get_url", ["salt://grail/nonexistent_scene", tgt])
        self.assertEqual(ret, False)

    @slowTest
    def test_get_url_to_dir(self):
        """
        cp.get_url with salt:// source
        """
        tgt = os.path.join(RUNTIME_VARS.TMP, "")
        self.run_function("cp.get_url", ["salt://grail/scene33", tgt])
        with salt.utils.files.fopen(tgt + "scene33", "r") as scene:
            data = salt.utils.stringutils.to_unicode(scene.read())
        self.assertIn("KNIGHT:  They're nervous, sire.", data)
        self.assertNotIn("bacon", data)

    @skipIf(
        salt.utils.platform.is_darwin() and six.PY2, "This test hangs on OS X on Py2"
    )
    @with_tempfile()
    @slowTest
    def test_get_url_https(self, tgt):
        """
        cp.get_url with https:// source given
        """
        self.run_function("cp.get_url", ["https://repo.saltstack.com/index.html", tgt])
        with salt.utils.files.fopen(tgt, "r") as instructions:
            data = salt.utils.stringutils.to_unicode(instructions.read())
        self.assertIn("Bootstrap", data)
        self.assertIn("Debian", data)
        self.assertIn("Windows", data)
        self.assertNotIn("AYBABTU", data)

    @skipIf(
        salt.utils.platform.is_darwin() and six.PY2, "This test hangs on OS X on Py2"
    )
    @slowTest
    def test_get_url_https_dest_empty(self):
        """
        cp.get_url with https:// source given and destination omitted.
        """
        ret = self.run_function("cp.get_url", ["https://repo.saltstack.com/index.html"])

        with salt.utils.files.fopen(ret, "r") as instructions:
            data = salt.utils.stringutils.to_unicode(instructions.read())
        self.assertIn("Bootstrap", data)
        self.assertIn("Debian", data)
        self.assertIn("Windows", data)
        self.assertNotIn("AYBABTU", data)

    @skipIf(
        salt.utils.platform.is_darwin() and six.PY2, "This test hangs on OS X on Py2"
    )
    @slowTest
    def test_get_url_https_no_dest(self):
        """
        cp.get_url with https:// source given and destination set as None
        """
        timeout = 500
        start = time.time()
        sleep = 5
        tgt = None
        while time.time() - start <= timeout:
            ret = self.run_function(
                "cp.get_url", ["https://repo.saltstack.com/index.html", tgt]
            )
            if ret.find("HTTP 599") == -1:
                break
            time.sleep(sleep)
        if ret.find("HTTP 599") != -1:
            raise Exception("https://repo.saltstack.com/index.html returned 599 error")
        self.assertIn("Bootstrap", ret)
        self.assertIn("Debian", ret)
        self.assertIn("Windows", ret)
        self.assertNotIn("AYBABTU", ret)

    @slowTest
    def test_get_url_file(self):
        """
        cp.get_url with file:// source given
        """
        tgt = ""
        src = os.path.join("file://", RUNTIME_VARS.FILES, "file", "base", "file.big")
        ret = self.run_function("cp.get_url", [src, tgt])
        with salt.utils.files.fopen(ret, "r") as scene:
            data = salt.utils.stringutils.to_unicode(scene.read())
        self.assertIn("KNIGHT:  They're nervous, sire.", data)
        self.assertNotIn("bacon", data)

    @slowTest
    def test_get_url_file_no_dest(self):
        """
        cp.get_url with file:// source given and destination set as None
        """
        tgt = None
        src = os.path.join("file://", RUNTIME_VARS.FILES, "file", "base", "file.big")
        ret = self.run_function("cp.get_url", [src, tgt])
        self.assertIn("KNIGHT:  They're nervous, sire.", ret)
        self.assertNotIn("bacon", ret)

    @with_tempfile()
    @slowTest
    def test_get_url_ftp(self, tgt):
        """
        cp.get_url with https:// source given
        """
        self.run_function(
            "cp.get_url",
            [
                "ftp://ftp.freebsd.org/pub/FreeBSD/releases/amd64/amd64/12.0-RELEASE/MANIFEST",
                tgt,
            ],
        )
        with salt.utils.files.fopen(tgt, "r") as instructions:
            data = salt.utils.stringutils.to_unicode(instructions.read())
        self.assertIn("Base system", data)

    # cp.get_file_str tests

    @slowTest
    def test_get_file_str_salt(self):
        """
        cp.get_file_str with salt:// source given
        """
        src = "salt://grail/scene33"
        ret = self.run_function("cp.get_file_str", [src])
        self.assertIn("KNIGHT:  They're nervous, sire.", ret)

    @slowTest
    def test_get_file_str_nonexistent_source(self):
        """
        cp.get_file_str with nonexistent salt:// source given
        """
        src = "salt://grail/nonexistent_scene"
        ret = self.run_function("cp.get_file_str", [src])
        self.assertEqual(ret, False)

    @skipIf(
        salt.utils.platform.is_darwin() and six.PY2, "This test hangs on OS X on Py2"
    )
    @slowTest
    def test_get_file_str_https(self):
        """
        cp.get_file_str with https:// source given
        """
        src = "https://repo.saltstack.com/index.html"
        ret = self.run_function("cp.get_file_str", [src])
        self.assertIn("Bootstrap", ret)
        self.assertIn("Debian", ret)
        self.assertIn("Windows", ret)
        self.assertNotIn("AYBABTU", ret)

    @slowTest
    def test_get_file_str_local(self):
        """
        cp.get_file_str with file:// source given
        """
        src = os.path.join("file://", RUNTIME_VARS.FILES, "file", "base", "file.big")
        ret = self.run_function("cp.get_file_str", [src])
        self.assertIn("KNIGHT:  They're nervous, sire.", ret)
        self.assertNotIn("bacon", ret)

    # caching tests

    @slowTest
    def test_cache_file(self):
        """
        cp.cache_file
        """
        ret = self.run_function("cp.cache_file", ["salt://grail/scene33"])
        with salt.utils.files.fopen(ret, "r") as scene:
            data = salt.utils.stringutils.to_unicode(scene.read())
        self.assertIn("KNIGHT:  They're nervous, sire.", data)
        self.assertNotIn("bacon", data)

    @slowTest
    def test_cache_files(self):
        """
        cp.cache_files
        """
        ret = self.run_function(
            "cp.cache_files", [["salt://grail/scene33", "salt://grail/36/scene"]]
        )
        for path in ret:
            with salt.utils.files.fopen(path, "r") as scene:
                data = salt.utils.stringutils.to_unicode(scene.read())
            self.assertIn("ARTHUR:", data)
            self.assertNotIn("bacon", data)

    @with_tempfile()
    @slowTest
    def test_cache_master(self, tgt):
        """
        cp.cache_master
        """
        ret = self.run_function("cp.cache_master", [tgt],)
        for path in ret:
            self.assertTrue(os.path.exists(path))

    @slowTest
    def test_cache_local_file(self):
        """
        cp.cache_local_file
        """
        src = os.path.join(RUNTIME_VARS.TMP, "random")
        with salt.utils.files.fopen(src, "w+") as fn_:
            fn_.write(salt.utils.stringutils.to_str("foo"))
        ret = self.run_function("cp.cache_local_file", [src])
        with salt.utils.files.fopen(ret, "r") as cp_:
            self.assertEqual(salt.utils.stringutils.to_unicode(cp_.read()), "foo")

    @skipIf(not salt.utils.path.which("nginx"), "nginx not installed")
    @skip_if_not_root
    @slowTest
    def test_cache_remote_file(self):
        """
        cp.cache_file
        """
        nginx_port = get_unused_localhost_port()
        url_prefix = "http://localhost:{0}/".format(nginx_port)
        temp_dir = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)
        self.addCleanup(shutil.rmtree, temp_dir, ignore_errors=True)
        nginx_root_dir = os.path.join(temp_dir, "root")
        nginx_conf_dir = os.path.join(temp_dir, "conf")
        nginx_conf = os.path.join(nginx_conf_dir, "nginx.conf")
        nginx_pidfile = os.path.join(nginx_conf_dir, "nginx.pid")
        file_contents = "Hello world!"

        for dirname in (nginx_root_dir, nginx_conf_dir):
            os.makedirs(dirname)

        # Write the temp file
        with salt.utils.files.fopen(
            os.path.join(nginx_root_dir, "actual_file"), "w"
        ) as fp_:
            fp_.write(salt.utils.stringutils.to_str(file_contents))

        # Write the nginx config
        with salt.utils.files.fopen(nginx_conf, "w") as fp_:
            fp_.write(
                textwrap.dedent(
                    salt.utils.stringutils.to_str(
                        """\
                user root;
                worker_processes 1;
                error_log {nginx_conf_dir}/server_error.log;
                pid {nginx_pidfile};

                events {{
                    worker_connections 1024;
                }}

                http {{
                    include       /etc/nginx/mime.types;
                    default_type  application/octet-stream;

                    access_log {nginx_conf_dir}/access.log;
                    error_log {nginx_conf_dir}/error.log;

                    server {{
                        listen {nginx_port} default_server;
                        server_name cachefile.local;
                        root {nginx_root_dir};

                        location ~ ^/301$ {{
                            return 301 /actual_file;
                        }}

                        location ~ ^/302$ {{
                            return 302 /actual_file;
                        }}
                    }}
                }}""".format(
                            **locals()
                        )
                    )
                )
            )

        self.run_function("cmd.run", [["nginx", "-c", nginx_conf]], python_shell=False)
        with salt.utils.files.fopen(nginx_pidfile) as fp_:
            nginx_pid = int(fp_.read().strip())
            nginx_proc = psutil.Process(pid=nginx_pid)
            self.addCleanup(nginx_proc.send_signal, signal.SIGQUIT)

        for code in ("", "301", "302"):
            url = url_prefix + (code or "actual_file")
            log.debug("attempting to cache %s", url)
            ret = self.run_function("cp.cache_file", [url])
            self.assertTrue(ret)
            with salt.utils.files.fopen(ret) as fp_:
                cached_contents = salt.utils.stringutils.to_unicode(fp_.read())
                self.assertEqual(cached_contents, file_contents)

    @slowTest
    def test_list_states(self):
        """
        cp.list_states
        """
        ret = self.run_function("cp.list_states",)
        self.assertIn("core", ret)
        self.assertIn("top", ret)

    @slowTest
    def test_list_minion(self):
        """
        cp.list_minion
        """
        self.run_function("cp.cache_file", ["salt://grail/scene33"])
        ret = self.run_function("cp.list_minion")
        found = False
        search = "grail/scene33"
        if salt.utils.platform.is_windows():
            search = r"grail\scene33"
        for path in ret:
            if search in path:
                found = True
                break
        self.assertTrue(found)

    @slowTest
    def test_is_cached(self):
        """
        cp.is_cached
        """
        self.run_function("cp.cache_file", ["salt://grail/scene33"])
        ret1 = self.run_function("cp.is_cached", ["salt://grail/scene33"])
        self.assertTrue(ret1)
        ret2 = self.run_function("cp.is_cached", ["salt://fasldkgj/poicxzbn"])
        self.assertFalse(ret2)

    @slowTest
    def test_hash_file(self):
        """
        cp.hash_file
        """
        sha256_hash = self.run_function("cp.hash_file", ["salt://grail/scene33"])
        path = self.run_function("cp.cache_file", ["salt://grail/scene33"])
        with salt.utils.files.fopen(path, "rb") as fn_:
            data = fn_.read()
            self.assertEqual(sha256_hash["hsum"], hashlib.sha256(data).hexdigest())

    @with_tempfile()
    @slowTest
    def test_get_file_from_env_predefined(self, tgt):
        """
        cp.get_file
        """
        tgt = os.path.join(RUNTIME_VARS.TMP, "cheese")
        try:
            self.run_function("cp.get_file", ["salt://cheese", tgt])
            with salt.utils.files.fopen(tgt, "r") as cheese:
                data = salt.utils.stringutils.to_unicode(cheese.read())
            self.assertIn("Gromit", data)
            self.assertNotIn("Comte", data)
        finally:
            os.unlink(tgt)

    @with_tempfile()
    @slowTest
    def test_get_file_from_env_in_url(self, tgt):
        tgt = os.path.join(RUNTIME_VARS.TMP, "cheese")
        try:
            self.run_function("cp.get_file", ["salt://cheese?saltenv=prod", tgt])
            with salt.utils.files.fopen(tgt, "r") as cheese:
                data = salt.utils.stringutils.to_unicode(cheese.read())
            self.assertIn("Gromit", data)
            self.assertIn("Comte", data)
        finally:
            os.unlink(tgt)

    @slowTest
    def test_push(self):
        log_to_xfer = os.path.join(RUNTIME_VARS.TMP, uuid.uuid4().hex)
        open(log_to_xfer, "w").close()  # pylint: disable=resource-leakage
        try:
            self.run_function("cp.push", [log_to_xfer])
            tgt_cache_file = os.path.join(
                RUNTIME_VARS.TMP,
                "master-minion-root",
                "cache",
                "minions",
                "minion",
                "files",
                RUNTIME_VARS.TMP,
                log_to_xfer,
            )
            self.assertTrue(
                os.path.isfile(tgt_cache_file), "File was not cached on the master"
            )
        finally:
            os.unlink(tgt_cache_file)

    @slowTest
    def test_envs(self):
        self.assertEqual(sorted(self.run_function("cp.envs")), sorted(["base", "prod"]))



# -*- coding: utf-8 -*-
"""
Initialize the engines system. This plugin system allows for
complex services to be encapsulated within the salt plugin environment
"""
# Import python libs
from __future__ import absolute_import, print_function, unicode_literals

import logging
import multiprocessing

# Import salt libs
import salt
import salt.loader
import salt.utils.platform
from salt.utils.process import SignalHandlingProcess

log = logging.getLogger(__name__)


def start_engines(opts, proc_mgr, proxy=None):
    """
    Fire up the configured engines!
    """
    utils = salt.loader.utils(opts, proxy=proxy)
    if opts["__role"] == "master":
        runners = salt.loader.runner(opts, utils=utils)
    else:
        runners = []
    funcs = salt.loader.minion_mods(opts, utils=utils, proxy=proxy)
    engines = salt.loader.engines(opts, funcs, runners, utils, proxy=proxy)

    engines_opt = opts.get("engines", [])
    if isinstance(engines_opt, dict):
        engines_opt = [{k: v} for k, v in engines_opt.items()]

    # Function references are not picklable. Windows needs to pickle when
    # spawning processes. On Windows, these will need to be recalculated
    # in the spawned child process.
    if salt.utils.platform.is_windows():
        runners = None
        utils = None
        funcs = None

    for engine in engines_opt:
        if isinstance(engine, dict):
            engine, engine_opts = next(iter(engine.items()))
        else:
            engine_opts = None
        engine_name = None
        if engine_opts is not None and "engine_module" in engine_opts:
            fun = "{0}.start".format(engine_opts["engine_module"])
            engine_name = engine
            del engine_opts["engine_module"]
        else:
            fun = "{0}.start".format(engine)
        if fun in engines:
            start_func = engines[fun]
            if engine_name:
                name = "{0}.Engine({1}-{2})".format(
                    __name__, start_func.__module__, engine_name
                )
            else:
                name = "{0}.Engine({1})".format(__name__, start_func.__module__)
            log.info("Starting Engine %s", name)
            proc_mgr.add_process(
                Engine, args=(opts, fun, engine_opts, funcs, runners, proxy), name=name
            )


class Engine(SignalHandlingProcess):
    """
    Execute the given engine in a new process
    """

    def __init__(self, opts, fun, config, funcs, runners, proxy, **kwargs):
        """
        Set up the process executor
        """
        super(Engine, self).__init__(**kwargs)
        self.opts = opts
        self.config = config
        self.fun = fun
        self.funcs = funcs
        self.runners = runners
        self.proxy = proxy

    # __setstate__ and __getstate__ are only used on Windows.
    # We do this so that __init__ will be invoked on Windows in the child
    # process so that a register_after_fork() equivalent will work on Windows.
    def __setstate__(self, state):
        self.__init__(
            state["opts"],
            state["fun"],
            state["config"],
            state["funcs"],
            state["runners"],
            state["proxy"],
            log_queue=state["log_queue"],
            log_queue_level=state["log_queue_level"],
        )

    def __getstate__(self):
        return {
            "opts": self.opts,
            "fun": self.fun,
            "config": self.config,
            "funcs": self.funcs,
            "runners": self.runners,
            "proxy": self.proxy,
            "log_queue": self.log_queue,
            "log_queue_level": self.log_queue_level,
        }

    def run(self):
        """
        Run the master service!
        """
        self.utils = salt.loader.utils(self.opts, proxy=self.proxy)
        if salt.utils.platform.is_windows():
            # Calculate function references since they can't be pickled.
            if self.opts["__role"] == "master":
                self.runners = salt.loader.runner(self.opts, utils=self.utils)
            else:
                self.runners = []
            self.funcs = salt.loader.minion_mods(
                self.opts, utils=self.utils, proxy=self.proxy
            )

        self.engine = salt.loader.engines(
            self.opts, self.funcs, self.runners, self.utils, proxy=self.proxy
        )
        kwargs = self.config or {}
        try:
            self.engine[self.fun](**kwargs)
        except Exception as exc:  # pylint: disable=broad-except
            log.critical(
                "Engine '%s' could not be started!",
                self.fun.split(".")[0],
                exc_info=True,
            )



# -*- coding: utf-8 -*-
"""
    :codeauthor: Rahul Handay <rahulha@saltstack.com>
"""

# Import Python Libs
from __future__ import absolute_import, print_function, unicode_literals

# Import Salt Libs
import salt.states.win_system as win_system

# Import Salt Testing Libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class WinSystemTestCase(TestCase, LoaderModuleMockMixin):
    """
        Validate the win_system state
    """

    def setup_loader_modules(self):
        return {win_system: {}}

    def test_computer_desc(self):
        """
            Test to manage the computer's description field
        """
        ret = {"name": "salt", "changes": {}, "result": True, "comment": ""}
        mock = MagicMock(side_effect=["salt", "stack", "stack"])
        with patch.dict(win_system.__salt__, {"system.get_computer_desc": mock}):
            ret.update({"comment": "Computer description" " already set to 'salt'"})
            self.assertDictEqual(win_system.computer_desc("salt"), ret)

            with patch.dict(win_system.__opts__, {"test": True}):
                ret.update(
                    {
                        "result": None,
                        "comment": "Computer description" " will be changed to 'salt'",
                    }
                )
                self.assertDictEqual(win_system.computer_desc("salt"), ret)

            with patch.dict(win_system.__opts__, {"test": False}):
                mock = MagicMock(return_value={"Computer Description": "nfs"})
                with patch.dict(
                    win_system.__salt__, {"system.set_computer_desc": mock}
                ):
                    ret.update(
                        {
                            "result": False,
                            "comment": "Unable to set"
                            " computer description to 'salt'",
                        }
                    )
                    self.assertDictEqual(win_system.computer_desc("salt"), ret)

    def test_computer_name(self):
        """
            Test to manage the computer's name
        """
        ret = {"name": "salt", "changes": {}, "result": True, "comment": ""}
        mock = MagicMock(return_value="salt")
        with patch.dict(win_system.__salt__, {"system.get_computer_name": mock}):
            mock = MagicMock(side_effect=[None, "SALT", "Stack", "stack"])
            with patch.dict(
                win_system.__salt__, {"system.get_pending_computer_name": mock}
            ):
                ret.update({"comment": "Computer name already set to 'salt'"})
                self.assertDictEqual(win_system.computer_name("salt"), ret)

                ret.update(
                    {
                        "comment": "The current computer name"
                        " is 'salt', but will be changed to 'salt' on"
                        " the next reboot"
                    }
                )
                self.assertDictEqual(win_system.computer_name("salt"), ret)

                with patch.dict(win_system.__opts__, {"test": True}):
                    ret.update(
                        {
                            "result": None,
                            "comment": "Computer name will" " be changed to 'salt'",
                        }
                    )
                    self.assertDictEqual(win_system.computer_name("salt"), ret)

                with patch.dict(win_system.__opts__, {"test": False}):
                    mock = MagicMock(return_value=False)
                    with patch.dict(
                        win_system.__salt__, {"system.set_computer_name": mock}
                    ):
                        ret.update(
                            {
                                "comment": "Unable to set computer name" " to 'salt'",
                                "result": False,
                            }
                        )
                        self.assertDictEqual(win_system.computer_name("salt"), ret)

    def test_hostname(self):
        ret = {"name": "salt", "changes": {}, "result": True, "comment": ""}

        mock = MagicMock(return_value="minion")
        with patch.dict(win_system.__salt__, {"system.get_hostname": mock}):
            mock = MagicMock(return_value=True)
            with patch.dict(win_system.__salt__, {"system.set_hostname": mock}):
                ret.update(
                    {
                        "comment": "The current hostname is 'minion', "
                        "but will be changed to 'salt' on the next reboot",
                        "changes": {"hostname": "salt"},
                    }
                )
                self.assertDictEqual(win_system.hostname("salt"), ret)

            mock = MagicMock(return_value=False)
            with patch.dict(win_system.__salt__, {"system.set_hostname": mock}):
                ret.update(
                    {
                        "comment": "Unable to set hostname",
                        "changes": {},
                        "result": False,
                    }
                )
                self.assertDictEqual(win_system.hostname("salt"), ret)

        mock = MagicMock(return_value="salt")
        with patch.dict(win_system.__salt__, {"system.get_hostname": mock}):
            ret.update(
                {
                    "comment": "Hostname is already set to 'salt'",
                    "changes": {},
                    "result": True,
                }
            )

            self.assertDictEqual(win_system.hostname("salt"), ret)

        mock = MagicMock(return_value="salt")
        with patch.dict(win_system.__salt__, {"system.get_hostname": mock}):
            ret.update(
                {
                    "name": "SALT",
                    "comment": "Hostname is already set to 'SALT'",
                    "changes": {},
                    "result": True,
                }
            )

            self.assertDictEqual(win_system.hostname("SALT"), ret)



# -*- coding: utf-8 -*-

# Import python libs
from __future__ import absolute_import, print_function, unicode_literals

# Imports Standards
import os

# Import salt libs
import salt.utils.platform
import salt.utils.user
from tests.support.mock import patch
from tests.support.runtests import this_user

# Import Salt Testing libs
from tests.support.unit import TestCase, skipIf

# Import Conditionals
try:
    import grp

    HAS_GRP = True
except ImportError:
    HAS_GRP = False

try:
    import pwd

    HAS_PWD = True
except ImportError:
    HAS_PWD = False


class TestUser(TestCase):
    @skipIf(HAS_GRP is False or HAS_PWD is False, "Module grp or pwd is missing")
    @skipIf(salt.utils.platform.is_windows(), "Module not available on Windows")
    def test_chugid_and_umask(self):

        running_user = this_user()
        running_group = grp.getgrgid(os.getgid()).gr_name

        gids = {30: "expectedgroup", 20: running_group}
        getgrnams = {
            "expectedgroup": grp.struct_group(
                ("expectedgroup", "*", 30, ["expecteduser"])
            ),
            running_group: grp.struct_group((running_group, "*", 20, [running_user])),
        }
        getpwnams = {
            "expecteduser": pwd.struct_passwd(
                ("expecteduser", "x", 30, 30, "-", "-", "-")
            ),
            running_user: pwd.struct_passwd((running_user, "x", 20, 20, "-", "-", "-")),
        }

        def getgrnam(group):
            return getgrnams[group]

        def getpwnam(user):
            return getpwnams[user]

        def getgrgid(gid):
            return getgrnams[gids[gid]]

        with patch("grp.getgrgid", getgrgid):
            with patch("grp.getgrnam", getgrnam):
                with patch("pwd.getpwnam", getpwnam):
                    with patch("salt.utils.user.chugid") as chugid_mock:
                        salt.utils.user.chugid_and_umask(
                            "expecteduser", umask=None, group=running_group
                        )
                        chugid_mock.assert_called_with("expecteduser", running_group)
                        salt.utils.user.chugid_and_umask(
                            "expecteduser", umask=None, group=None
                        )
                        chugid_mock.assert_called_with("expecteduser", "expectedgroup")



# -*- coding: utf-8 -*-
"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

from __future__ import absolute_import, print_function, unicode_literals

import logging
import os

import salt.modules.schedule as schedule
from salt.utils.event import SaltEvent
from tests.support.helpers import slowTest
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import TestCase

log = logging.getLogger(__name__)

JOB1 = {
    "function": "test.ping",
    "maxrunning": 1,
    "name": "job1",
    "jid_include": True,
    "enabled": True,
}


class ScheduleTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.schedule
    """

    @classmethod
    def setUpClass(cls):
        cls.sock_dir = os.path.join(RUNTIME_VARS.TMP, "test-socks")

    def setup_loader_modules(self):
        return {schedule: {}}

    # 'purge' function tests: 1

    @slowTest
    def test_purge(self):
        """
        Test if it purge all the jobs currently scheduled on the minion.
        """
        with patch.dict(schedule.__opts__, {"schedule": {}, "sock_dir": self.sock_dir}):
            mock = MagicMock(return_value=True)
            with patch.dict(schedule.__salt__, {"event.fire": mock}):
                _ret_value = {"complete": True, "schedule": {}}
                with patch.object(SaltEvent, "get_event", return_value=_ret_value):
                    self.assertDictEqual(
                        schedule.purge(),
                        {
                            "comment": ["Deleted job: schedule from schedule."],
                            "result": True,
                        },
                    )

    # 'delete' function tests: 1

    @slowTest
    def test_delete(self):
        """
        Test if it delete a job from the minion's schedule.
        """
        with patch.dict(schedule.__opts__, {"schedule": {}, "sock_dir": self.sock_dir}):
            mock = MagicMock(return_value=True)
            with patch.dict(schedule.__salt__, {"event.fire": mock}):
                _ret_value = {"complete": True, "schedule": {}}
                with patch.object(SaltEvent, "get_event", return_value=_ret_value):
                    self.assertDictEqual(
                        schedule.delete("job1"),
                        {"comment": "Job job1 does not exist.", "result": False},
                    )

    # 'build_schedule_item' function tests: 1

    def test_build_schedule_item(self):
        """
        Test if it build a schedule job.
        """
        comment = (
            'Unable to use "seconds", "minutes", "hours", '
            'or "days" with "when" or "cron" options.'
        )
        comment1 = 'Unable to use "when" and "cron" ' "options together.  Ignoring."
        with patch.dict(schedule.__opts__, {"job1": {}}):
            self.assertDictEqual(
                schedule.build_schedule_item(""),
                {"comment": "Job name is required.", "result": False},
            )

            self.assertDictEqual(
                schedule.build_schedule_item("job1", function="test.ping"),
                {
                    "function": "test.ping",
                    "maxrunning": 1,
                    "name": "job1",
                    "jid_include": True,
                    "enabled": True,
                },
            )

            self.assertDictEqual(
                schedule.build_schedule_item(
                    "job1", function="test.ping", seconds=3600, when="2400"
                ),
                {"comment": comment, "result": False},
            )

            self.assertDictEqual(
                schedule.build_schedule_item(
                    "job1", function="test.ping", when="2400", cron="2"
                ),
                {"comment": comment1, "result": False},
            )

    # 'build_schedule_item_invalid_when' function tests: 1

    def test_build_schedule_item_invalid_when(self):
        """
        Test if it build a schedule job.
        """
        comment = 'Schedule item garbage for "when" in invalid.'
        with patch.dict(schedule.__opts__, {"job1": {}}):
            self.assertDictEqual(
                schedule.build_schedule_item(
                    "job1", function="test.ping", when="garbage"
                ),
                {"comment": comment, "result": False},
            )

    # 'add' function tests: 1

    @slowTest
    def test_add(self):
        """
        Test if it add a job to the schedule.
        """
        comm1 = "Job job1 already exists in schedule."
        comm2 = (
            'Error: Unable to use "seconds", "minutes", "hours", '
            'or "days" with "when" or "cron" options.'
        )
        comm3 = 'Unable to use "when" and "cron" options together.  Ignoring.'
        comm4 = "Job: job2 would be added to schedule."
        with patch.dict(
            schedule.__opts__, {"schedule": {"job1": "salt"}, "sock_dir": self.sock_dir}
        ):
            mock = MagicMock(return_value=True)
            with patch.dict(schedule.__salt__, {"event.fire": mock}):
                _ret_value = {"complete": True, "schedule": {"job1": {"salt": "salt"}}}
                with patch.object(SaltEvent, "get_event", return_value=_ret_value):
                    self.assertDictEqual(
                        schedule.add("job1"), {"comment": comm1, "result": False}
                    )

                _ret_value = {"complete": True, "schedule": {}}
                with patch.object(SaltEvent, "get_event", return_value=_ret_value):
                    self.assertDictEqual(
                        schedule.add(
                            "job2", function="test.ping", seconds=3600, when="2400"
                        ),
                        {"comment": comm2, "result": False},
                    )

                _ret_value = {"complete": True, "schedule": {}}
                with patch.object(SaltEvent, "get_event", return_value=_ret_value):
                    self.assertDictEqual(
                        schedule.add(
                            "job2", function="test.ping", when="2400", cron="2"
                        ),
                        {"comment": comm3, "result": False},
                    )
                _ret_value = {"complete": True, "schedule": {}}
                with patch.object(SaltEvent, "get_event", return_value=_ret_value):
                    self.assertDictEqual(
                        schedule.add("job2", function="test.ping", test=True),
                        {"comment": comm4, "result": True},
                    )

    # 'run_job' function tests: 1

    @slowTest
    def test_run_job(self):
        """
        Test if it run a scheduled job on the minion immediately.
        """
        with patch.dict(
            schedule.__opts__, {"schedule": {"job1": JOB1}, "sock_dir": self.sock_dir}
        ):
            mock = MagicMock(return_value=True)
            with patch.dict(schedule.__salt__, {"event.fire": mock}):
                _ret_value = {"complete": True, "schedule": {"job1": JOB1}}
                with patch.object(SaltEvent, "get_event", return_value=_ret_value):
                    self.assertDictEqual(
                        schedule.run_job("job1"),
                        {"comment": "Scheduling Job job1 on minion.", "result": True},
                    )

    # 'enable_job' function tests: 1

    @slowTest
    def test_enable_job(self):
        """
        Test if it enable a job in the minion's schedule.
        """
        with patch.dict(schedule.__opts__, {"schedule": {}, "sock_dir": self.sock_dir}):
            mock = MagicMock(return_value=True)
            with patch.dict(schedule.__salt__, {"event.fire": mock}):
                _ret_value = {"complete": True, "schedule": {}}
                with patch.object(SaltEvent, "get_event", return_value=_ret_value):
                    self.assertDictEqual(
                        schedule.enable_job("job1"),
                        {"comment": "Job job1 does not exist.", "result": False},
                    )

    # 'disable_job' function tests: 1

    @slowTest
    def test_disable_job(self):
        """
        Test if it disable a job in the minion's schedule.
        """
        with patch.dict(schedule.__opts__, {"schedule": {}, "sock_dir": self.sock_dir}):
            mock = MagicMock(return_value=True)
            with patch.dict(schedule.__salt__, {"event.fire": mock}):
                _ret_value = {"complete": True, "schedule": {}}
                with patch.object(SaltEvent, "get_event", return_value=_ret_value):
                    self.assertDictEqual(
                        schedule.disable_job("job1"),
                        {"comment": "Job job1 does not exist.", "result": False},
                    )

    # 'save' function tests: 1

    @slowTest
    def test_save(self):
        """
        Test if it save all scheduled jobs on the minion.
        """
        comm1 = "Schedule (non-pillar items) saved."
        with patch.dict(
            schedule.__opts__,
            {"schedule": {}, "default_include": "/tmp", "sock_dir": self.sock_dir},
        ):

            mock = MagicMock(return_value=True)
            with patch.dict(schedule.__salt__, {"event.fire": mock}):
                _ret_value = {"complete": True, "schedule": {}}
                with patch.object(SaltEvent, "get_event", return_value=_ret_value):
                    self.assertDictEqual(
                        schedule.save(), {"comment": comm1, "result": True}
                    )

    # 'enable' function tests: 1

    def test_enable(self):
        """
        Test if it enable all scheduled jobs on the minion.
        """
        self.assertDictEqual(
            schedule.enable(test=True),
            {"comment": "Schedule would be enabled.", "result": True},
        )

    # 'disable' function tests: 1

    def test_disable(self):
        """
        Test if it disable all scheduled jobs on the minion.
        """
        self.assertDictEqual(
            schedule.disable(test=True),
            {"comment": "Schedule would be disabled.", "result": True},
        )

    # 'move' function tests: 1

    @slowTest
    def test_move(self):
        """
        Test if it move scheduled job to another minion or minions.
        """
        comm1 = "no servers answered the published schedule.add command"
        comm2 = "the following minions return False"
        comm3 = "Moved Job job1 from schedule."
        with patch.dict(
            schedule.__opts__, {"schedule": {"job1": JOB1}, "sock_dir": self.sock_dir}
        ):
            mock = MagicMock(return_value=True)
            with patch.dict(schedule.__salt__, {"event.fire": mock}):
                _ret_value = {"complete": True, "schedule": {"job1": JOB1}}
                with patch.object(SaltEvent, "get_event", return_value=_ret_value):
                    mock = MagicMock(return_value={})
                    with patch.dict(schedule.__salt__, {"publish.publish": mock}):
                        self.assertDictEqual(
                            schedule.move("job1", "minion1"),
                            {"comment": comm1, "result": True},
                        )

                    mock = MagicMock(return_value={"minion1": ""})
                    with patch.dict(schedule.__salt__, {"publish.publish": mock}):
                        self.assertDictEqual(
                            schedule.move("job1", "minion1"),
                            {"comment": comm2, "minions": ["minion1"], "result": True},
                        )

                    mock = MagicMock(return_value={"minion1": "job1"})
                    with patch.dict(schedule.__salt__, {"publish.publish": mock}):
                        mock = MagicMock(return_value=True)
                        with patch.dict(schedule.__salt__, {"event.fire": mock}):
                            self.assertDictEqual(
                                schedule.move("job1", "minion1"),
                                {
                                    "comment": comm3,
                                    "minions": ["minion1"],
                                    "result": True,
                                },
                            )

                    self.assertDictEqual(
                        schedule.move("job3", "minion1"),
                        {"comment": "Job job3 does not exist.", "result": False},
                    )

        mock = MagicMock(side_effect=[{}, {"job1": {}}])
        with patch.dict(
            schedule.__opts__, {"schedule": mock, "sock_dir": self.sock_dir}
        ):
            mock = MagicMock(return_value=True)
            with patch.dict(schedule.__salt__, {"event.fire": mock}):
                _ret_value = {"complete": True, "schedule": {"job1": JOB1}}
                with patch.object(SaltEvent, "get_event", return_value=_ret_value):
                    with patch.dict(schedule.__pillar__, {"schedule": {"job1": JOB1}}):
                        mock = MagicMock(return_value={})
                        with patch.dict(schedule.__salt__, {"publish.publish": mock}):
                            self.assertDictEqual(
                                schedule.move("job1", "minion1"),
                                {"comment": comm1, "result": True},
                            )

                        mock = MagicMock(return_value={"minion1": ""})
                        with patch.dict(schedule.__salt__, {"publish.publish": mock}):
                            self.assertDictEqual(
                                schedule.move("job1", "minion1"),
                                {
                                    "comment": comm2,
                                    "minions": ["minion1"],
                                    "result": True,
                                },
                            )

                        mock = MagicMock(return_value={"minion1": "job1"})
                        with patch.dict(schedule.__salt__, {"publish.publish": mock}):
                            mock = MagicMock(return_value=True)
                            with patch.dict(schedule.__salt__, {"event.fire": mock}):
                                self.assertDictEqual(
                                    schedule.move("job1", "minion1"),
                                    {
                                        "comment": comm3,
                                        "minions": ["minion1"],
                                        "result": True,
                                    },
                                )

    # 'copy' function tests: 1

    @slowTest
    def test_copy(self):
        """
        Test if it copy scheduled job to another minion or minions.
        """
        comm1 = "no servers answered the published schedule.add command"
        comm2 = "the following minions return False"
        comm3 = "Copied Job job1 from schedule to minion(s)."
        with patch.dict(
            schedule.__opts__, {"schedule": {"job1": JOB1}, "sock_dir": self.sock_dir}
        ):
            mock = MagicMock(return_value=True)
            with patch.dict(schedule.__salt__, {"event.fire": mock}):
                _ret_value = {"complete": True, "schedule": {"job1": {"job1": JOB1}}}
                with patch.object(SaltEvent, "get_event", return_value=_ret_value):
                    mock = MagicMock(return_value={})
                    with patch.dict(schedule.__salt__, {"publish.publish": mock}):
                        self.assertDictEqual(
                            schedule.copy("job1", "minion1"),
                            {"comment": comm1, "result": True},
                        )

                    mock = MagicMock(return_value={"minion1": ""})
                    with patch.dict(schedule.__salt__, {"publish.publish": mock}):
                        self.assertDictEqual(
                            schedule.copy("job1", "minion1"),
                            {"comment": comm2, "minions": ["minion1"], "result": True},
                        )

                    mock = MagicMock(return_value={"minion1": "job1"})
                    with patch.dict(schedule.__salt__, {"publish.publish": mock}):
                        mock = MagicMock(return_value=True)
                        with patch.dict(schedule.__salt__, {"event.fire": mock}):
                            self.assertDictEqual(
                                schedule.copy("job1", "minion1"),
                                {
                                    "comment": comm3,
                                    "minions": ["minion1"],
                                    "result": True,
                                },
                            )

                    self.assertDictEqual(
                        schedule.copy("job3", "minion1"),
                        {"comment": "Job job3 does not exist.", "result": False},
                    )

        mock = MagicMock(side_effect=[{}, {"job1": {}}])
        with patch.dict(
            schedule.__opts__, {"schedule": mock, "sock_dir": self.sock_dir}
        ):
            with patch.dict(schedule.__pillar__, {"schedule": {"job1": JOB1}}):
                mock = MagicMock(return_value=True)
                with patch.dict(schedule.__salt__, {"event.fire": mock}):
                    _ret_value = {
                        "complete": True,
                        "schedule": {"job1": {"job1": JOB1}},
                    }
                    with patch.object(SaltEvent, "get_event", return_value=_ret_value):

                        mock = MagicMock(return_value={})
                        with patch.dict(schedule.__salt__, {"publish.publish": mock}):
                            self.assertDictEqual(
                                schedule.copy("job1", "minion1"),
                                {"comment": comm1, "result": True},
                            )

                        mock = MagicMock(return_value={"minion1": ""})
                        with patch.dict(schedule.__salt__, {"publish.publish": mock}):
                            self.assertDictEqual(
                                schedule.copy("job1", "minion1"),
                                {
                                    "comment": comm2,
                                    "minions": ["minion1"],
                                    "result": True,
                                },
                            )

                        mock = MagicMock(return_value={"minion1": "job1"})
                        with patch.dict(schedule.__salt__, {"publish.publish": mock}):
                            mock = MagicMock(return_value=True)
                            with patch.dict(schedule.__salt__, {"event.fire": mock}):
                                self.assertDictEqual(
                                    schedule.copy("job1", "minion1"),
                                    {
                                        "comment": comm3,
                                        "minions": ["minion1"],
                                        "result": True,
                                    },
                                )

    # 'modify' function tests: 1

    @slowTest
    def test_modify(self):
        """
        Test if modifying job to the schedule.
        """
        job1 = {"function": "salt", "seconds": 3600}

        comm1 = "Modified job: job1 in schedule."
        diff1 = (
            "--- \n+++ \n@@ -1,3 +1,6 @@\n "
            "enabled:True\n function:salt\n"
            "-seconds:3600\n+jid_include:True\n"
            "+maxrunning:1\n+name:job1\n"
            "+seconds:60\n"
        )

        diff4 = (
            "--- \n+++ \n@@ -1,3 +1,5 @@\n "
            "enabled:True\n-function:salt\n"
            "-seconds:3600\n+function:test.version\n"
            "+jid_include:True\n+maxrunning:1\n"
            "+name:job1\n"
        )

        expected1 = {"comment": comm1, "changes": {"diff": diff1}, "result": True}

        comm2 = (
            'Error: Unable to use "seconds", "minutes", "hours", '
            'or "days" with "when" option.'
        )
        expected2 = {"comment": comm2, "changes": {}, "result": False}

        comm3 = 'Unable to use "when" and "cron" options together.  Ignoring.'
        expected3 = {"comment": comm3, "changes": {}, "result": False}

        comm4 = "Job: job1 would be modified in schedule."
        expected4 = {"comment": comm4, "changes": {"diff": diff4}, "result": True}

        comm5 = "Job job2 does not exist in schedule."
        expected5 = {"comment": comm5, "changes": {}, "result": False}

        with patch.dict(
            schedule.__opts__, {"schedule": {"job1": job1}, "sock_dir": self.sock_dir}
        ):
            mock = MagicMock(return_value=True)
            with patch.dict(schedule.__salt__, {"event.fire": mock}):
                _ret_value = {"complete": True, "schedule": {"job1": job1}}
                with patch.object(SaltEvent, "get_event", return_value=_ret_value):
                    ret = schedule.modify("job1", seconds="60")
                    self.assertDictEqual(ret, expected1)

                _ret_value = {"complete": True, "schedule": {"job1": job1}}
                with patch.object(SaltEvent, "get_event", return_value=_ret_value):
                    ret = schedule.modify(
                        "job1", function="test.ping", seconds=3600, when="2400"
                    )
                    self.assertDictEqual(ret, expected2)

                _ret_value = {"complete": True, "schedule": {"job1": job1}}
                with patch.object(SaltEvent, "get_event", return_value=_ret_value):
                    ret = schedule.modify(
                        "job1", function="test.ping", when="2400", cron="2"
                    )
                    self.assertDictEqual(ret, expected3)

                _ret_value = {"complete": True, "schedule": {"job1": job1}}
                with patch.object(SaltEvent, "get_event", return_value=_ret_value):
                    ret = schedule.modify("job1", function="test.version", test=True)
                    self.assertDictEqual(ret, expected4)

                _ret_value = {"complete": True, "schedule": {}}
                with patch.object(SaltEvent, "get_event", return_value=_ret_value):
                    ret = schedule.modify("job2", function="test.version", test=True)
                    self.assertDictEqual(ret, expected5)

    # 'is_enabled' function tests: 1

    def test_is_enabled(self):
        """
        Test is_enabled
        """
        job1 = {"function": "salt", "seconds": 3600}

        comm1 = "Modified job: job1 in schedule."

        mock_schedule = {"enabled": True, "job1": job1}

        mock_lst = MagicMock(return_value=mock_schedule)

        with patch.dict(
            schedule.__opts__, {"schedule": {"job1": job1}, "sock_dir": self.sock_dir}
        ):
            mock = MagicMock(return_value=True)
            with patch.dict(
                schedule.__salt__, {"event.fire": mock, "schedule.list": mock_lst}
            ):
                _ret_value = {"complete": True, "schedule": {"job1": job1}}
                with patch.object(SaltEvent, "get_event", return_value=_ret_value):
                    ret = schedule.is_enabled("job1")
                    self.assertDictEqual(ret, job1)

                    ret = schedule.is_enabled()
                    self.assertEqual(ret, True)



# -*- coding: utf-8 -*-
"""
This is the default grains matcher function.
"""
from __future__ import absolute_import, print_function, unicode_literals

import logging

import salt.utils.data  # pylint: disable=3rd-party-module-not-gated
from salt.defaults import (  # pylint: disable=3rd-party-module-not-gated
    DEFAULT_TARGET_DELIM,
)

log = logging.getLogger(__name__)


def match(tgt, delimiter=DEFAULT_TARGET_DELIM, opts=None):
    """
    Reads in the grains glob match
    """
    if not opts:
        opts = __opts__

    log.debug("grains target: %s", tgt)
    if delimiter not in tgt:
        log.error(
            "Got insufficient arguments for grains match " "statement from master"
        )
        return False

    return salt.utils.data.subdict_match(opts["grains"], tgt, delimiter=delimiter)



# -*- coding: utf-8 -*-
"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""
# Import Python libs
from __future__ import absolute_import, print_function, unicode_literals

# Import Salt Libs
import salt.states.ntp as ntp

# Import Salt Testing Libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class NtpTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.states.ntp
    """

    def setup_loader_modules(self):
        return {ntp: {}}

    # 'managed' function tests: 1

    def test_managed(self):
        """
        Test to manage NTP servers.
        """
        name = "coffee-script"

        ret = {"name": name, "result": False, "comment": "", "changes": {}}

        mock_lst = MagicMock(return_value=[])
        with patch.dict(
            ntp.__salt__, {"ntp.get_servers": mock_lst, "ntp.set_servers": mock_lst}
        ):
            comt = "NTP servers already configured as specified"
            ret.update({"comment": comt, "result": True})
            self.assertDictEqual(ntp.managed(name, []), ret)

            with patch.dict(ntp.__opts__, {"test": True}):
                comt = "NTP servers will be updated to: coffee-script"
                ret.update({"comment": comt, "result": None})
                self.assertDictEqual(ntp.managed(name, [name]), ret)

            with patch.dict(ntp.__opts__, {"test": False}):
                comt = "Failed to update NTP servers"
                ret.update({"comment": comt, "result": False})
                self.assertDictEqual(ntp.managed(name, [name]), ret)



# coding: utf-8

# Python libs
from __future__ import absolute_import

import logging
import os
import shutil
import tempfile

# Salt libs
import salt.utils.files
from salt.beacons import inotify
from tests.support.mixins import LoaderModuleMockMixin

# Salt testing libs
from tests.support.unit import TestCase, skipIf

# Third-party libs
try:
    import pyinotify  # pylint: disable=unused-import

    HAS_PYINOTIFY = True
except ImportError:
    HAS_PYINOTIFY = False


log = logging.getLogger(__name__)


@skipIf(not HAS_PYINOTIFY, "pyinotify is not available")
class INotifyBeaconTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test case for salt.beacons.inotify
    """

    def setup_loader_modules(self):
        return {inotify: {}}

    def setUp(self):
        self.tmpdir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.tmpdir, ignore_errors=True)

    def test_non_list_config(self):
        config = {}

        ret = inotify.validate(config)

        self.assertEqual(
            ret, (False, "Configuration for inotify beacon must be a list.")
        )

    def test_empty_config(self):
        config = [{}]
        ret = inotify.validate(config)
        _expected = (False, "Configuration for inotify beacon must include files.")
        self.assertEqual(ret, _expected)

    def test_files_none_config(self):
        config = [{"files": None}]
        ret = inotify.validate(config)
        _expected = (
            False,
            "Configuration for inotify beacon invalid, files must be a dict.",
        )
        self.assertEqual(ret, _expected)

    def test_files_list_config(self):
        config = [{"files": [{u"/importantfile": {u"mask": [u"modify"]}}]}]
        ret = inotify.validate(config)
        _expected = (
            False,
            "Configuration for inotify beacon invalid, files must be a dict.",
        )
        self.assertEqual(ret, _expected)

    def test_file_open(self):
        path = os.path.realpath(__file__)
        config = [{"files": {path: {"mask": ["open"]}}}]
        ret = inotify.validate(config)
        self.assertEqual(ret, (True, "Valid beacon configuration"))

        ret = inotify.beacon(config)
        self.assertEqual(ret, [])

        with salt.utils.files.fopen(path, "r") as f:
            pass
        ret = inotify.beacon(config)
        self.assertEqual(len(ret), 1)
        self.assertEqual(ret[0]["path"], path)
        self.assertEqual(ret[0]["change"], "IN_OPEN")

    def test_dir_no_auto_add(self):
        config = [{"files": {self.tmpdir: {"mask": ["create"]}}}]
        ret = inotify.validate(config)
        self.assertEqual(ret, (True, "Valid beacon configuration"))

        ret = inotify.beacon(config)
        self.assertEqual(ret, [])
        fp = os.path.join(self.tmpdir, "tmpfile")
        with salt.utils.files.fopen(fp, "w") as f:
            pass
        ret = inotify.beacon(config)
        self.assertEqual(len(ret), 1)
        self.assertEqual(ret[0]["path"], fp)
        self.assertEqual(ret[0]["change"], "IN_CREATE")
        with salt.utils.files.fopen(fp, "r") as f:
            pass
        ret = inotify.beacon(config)
        self.assertEqual(ret, [])

    def test_dir_auto_add(self):
        config = [
            {"files": {self.tmpdir: {"mask": ["create", "open"], "auto_add": True}}}
        ]
        ret = inotify.validate(config)
        self.assertEqual(ret, (True, "Valid beacon configuration"))

        ret = inotify.beacon(config)
        self.assertEqual(ret, [])
        fp = os.path.join(self.tmpdir, "tmpfile")
        with salt.utils.files.fopen(fp, "w") as f:
            pass
        ret = inotify.beacon(config)
        self.assertEqual(len(ret), 2)
        self.assertEqual(ret[0]["path"], fp)
        self.assertEqual(ret[0]["change"], "IN_CREATE")
        self.assertEqual(ret[1]["path"], fp)
        self.assertEqual(ret[1]["change"], "IN_OPEN")
        with salt.utils.files.fopen(fp, "r") as f:
            pass
        ret = inotify.beacon(config)
        self.assertEqual(len(ret), 1)
        self.assertEqual(ret[0]["path"], fp)
        self.assertEqual(ret[0]["change"], "IN_OPEN")

    def test_dir_recurse(self):
        dp1 = os.path.join(self.tmpdir, "subdir1")
        os.mkdir(dp1)
        dp2 = os.path.join(dp1, "subdir2")
        os.mkdir(dp2)
        fp = os.path.join(dp2, "tmpfile")
        with salt.utils.files.fopen(fp, "w") as f:
            pass
        config = [{"files": {self.tmpdir: {"mask": ["open"], "recurse": True}}}]
        ret = inotify.validate(config)
        self.assertEqual(ret, (True, "Valid beacon configuration"))

        ret = inotify.beacon(config)
        self.assertEqual(ret, [])
        with salt.utils.files.fopen(fp) as f:
            pass
        ret = inotify.beacon(config)
        self.assertEqual(len(ret), 3)
        self.assertEqual(ret[0]["path"], dp1)
        self.assertEqual(ret[0]["change"], "IN_OPEN|IN_ISDIR")
        self.assertEqual(ret[1]["path"], dp2)
        self.assertEqual(ret[1]["change"], "IN_OPEN|IN_ISDIR")
        self.assertEqual(ret[2]["path"], fp)
        self.assertEqual(ret[2]["change"], "IN_OPEN")

    def test_dir_recurse_auto_add(self):
        dp1 = os.path.join(self.tmpdir, "subdir1")
        os.mkdir(dp1)
        config = [
            {
                "files": {
                    self.tmpdir: {
                        "mask": ["create", "delete"],
                        "recurse": True,
                        "auto_add": True,
                    }
                }
            }
        ]
        ret = inotify.validate(config)
        self.assertEqual(ret, (True, "Valid beacon configuration"))

        ret = inotify.beacon(config)
        self.assertEqual(ret, [])
        dp2 = os.path.join(dp1, "subdir2")
        os.mkdir(dp2)
        ret = inotify.beacon(config)
        self.assertEqual(len(ret), 1)
        self.assertEqual(ret[0]["path"], dp2)
        self.assertEqual(ret[0]["change"], "IN_CREATE|IN_ISDIR")
        fp = os.path.join(dp2, "tmpfile")
        with salt.utils.files.fopen(fp, "w") as f:
            pass
        ret = inotify.beacon(config)
        self.assertEqual(len(ret), 1)
        self.assertEqual(ret[0]["path"], fp)
        self.assertEqual(ret[0]["change"], "IN_CREATE")
        os.remove(fp)
        ret = inotify.beacon(config)
        self.assertEqual(len(ret), 1)
        self.assertEqual(ret[0]["path"], fp)
        self.assertEqual(ret[0]["change"], "IN_DELETE")

    def test_multi_files_exclude(self):
        dp1 = os.path.join(self.tmpdir, "subdir1")
        dp2 = os.path.join(self.tmpdir, "subdir2")
        os.mkdir(dp1)
        os.mkdir(dp2)
        _exclude1 = "{0}/subdir1/*tmpfile*$".format(self.tmpdir)
        _exclude2 = "{0}/subdir2/*filetmp*$".format(self.tmpdir)
        config = [
            {
                "files": {
                    dp1: {
                        "mask": ["create", "delete"],
                        "recurse": True,
                        "exclude": [{_exclude1: {"regex": True}}],
                        "auto_add": True,
                    }
                }
            },
            {
                "files": {
                    dp2: {
                        "mask": ["create", "delete"],
                        "recurse": True,
                        "exclude": [{_exclude2: {"regex": True}}],
                        "auto_add": True,
                    }
                }
            },
        ]
        ret = inotify.validate(config)
        self.assertEqual(ret, (True, "Valid beacon configuration"))

        fp = os.path.join(dp1, "tmpfile")
        with salt.utils.files.fopen(fp, "w") as f:
            pass
        ret = inotify.beacon(config)
        self.assertEqual(len(ret), 0)
        os.remove(fp)
        ret = inotify.beacon(config)
        self.assertEqual(len(ret), 0)

        fp = os.path.join(dp2, "tmpfile")
        with salt.utils.files.fopen(fp, "w") as f:
            pass
        ret = inotify.beacon(config)
        self.assertEqual(len(ret), 1)
        self.assertEqual(ret[0]["path"], fp)
        self.assertEqual(ret[0]["change"], "IN_CREATE")
        os.remove(fp)
        ret = inotify.beacon(config)
        self.assertEqual(len(ret), 1)
        self.assertEqual(ret[0]["path"], fp)
        self.assertEqual(ret[0]["change"], "IN_DELETE")



# -*- coding: utf-8 -*-
"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""
# Import Python libs
from __future__ import absolute_import, print_function, unicode_literals

# Import Salt Libs
import salt.states.openstack_config as openstack_config
from salt.exceptions import CommandExecutionError

# Import Salt Testing Libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class OpenstackConfigTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.states.openstack_config
    """

    def setup_loader_modules(self):
        return {openstack_config: {"__opts__": {"test": False}}}

    # 'present' function tests: 1

    def test_present(self):
        """
        Test to ensure a value is set in an OpenStack configuration file.
        """
        name = "salt"
        filename = "/tmp/salt"
        section = "A"
        value = "SALT"

        ret = {"name": name, "result": False, "comment": "", "changes": {}}

        mock_lst = MagicMock(side_effect=[value, CommandExecutionError, "A"])
        mock_t = MagicMock(return_value=True)
        with patch.dict(
            openstack_config.__salt__,
            {"openstack_config.get": mock_lst, "openstack_config.set": mock_t},
        ):
            comt = "The value is already set to the correct value"
            ret.update({"comment": comt, "result": True})
            self.assertDictEqual(
                openstack_config.present(name, filename, section, value), ret
            )

            self.assertRaises(
                CommandExecutionError,
                openstack_config.present,
                name,
                filename,
                section,
                value,
            )

            comt = "The value has been updated"
            ret.update({"comment": comt, "changes": {"Value": "Updated"}})
            self.assertDictEqual(
                openstack_config.present(name, filename, section, value), ret
            )

    # 'absent' function tests: 1

    def test_absent(self):
        """
        Test to ensure a value is not set in an OpenStack configuration file.
        """
        name = "salt"
        filename = "/tmp/salt"
        section = "A"

        ret = {"name": name, "result": False, "comment": "", "changes": {}}

        mock_lst = MagicMock(
            side_effect=[
                CommandExecutionError("parameter not found:"),
                CommandExecutionError,
                "A",
            ]
        )
        mock_t = MagicMock(return_value=True)
        with patch.dict(
            openstack_config.__salt__,
            {"openstack_config.get": mock_lst, "openstack_config.delete": mock_t},
        ):
            comt = "The value is already absent"
            ret.update({"comment": comt, "result": True})
            self.assertDictEqual(openstack_config.absent(name, filename, section), ret)

            self.assertRaises(
                CommandExecutionError, openstack_config.absent, name, filename, section
            )

            comt = "The value has been deleted"
            ret.update({"comment": comt, "changes": {"Value": "Deleted"}})
            self.assertDictEqual(openstack_config.absent(name, filename, section), ret)



# -*- coding: utf-8 -*-

from __future__ import absolute_import

import logging

from tests.support.mock import MagicMock, patch
from tests.support.unit import skipIf
from tests.unit.utils.scheduler.base import SchedulerTestsBase

try:
    import dateutil.parser

    HAS_DATEUTIL_PARSER = True
except ImportError:
    HAS_DATEUTIL_PARSER = False

try:
    import croniter  # pylint: disable=unused-import

    HAS_CRONITER = True
except ImportError:
    HAS_CRONITER = False

log = logging.getLogger(__name__)


@skipIf(
    HAS_DATEUTIL_PARSER is False, "The 'dateutil.parser' library is not available",
)
class SchedulerErrorTest(SchedulerTestsBase):
    def setUp(self):
        super(SchedulerErrorTest, self).setUp()
        self.schedule.opts["loop_interval"] = 1

        self.schedule.opts["grains"]["whens"] = {"tea time": "11/29/2017 12:00pm"}

    @skipIf(not HAS_CRONITER, "Cannot find croniter python module")
    def test_eval_cron_invalid(self):
        """
        verify that scheduled job runs
        """
        job = {"schedule": {"job1": {"function": "test.ping", "cron": "0 16 29 13 *"}}}

        # Add the job to the scheduler
        self.schedule.opts.update(job)

        run_time = dateutil.parser.parse("11/29/2017 4:00pm")
        with patch("croniter.croniter.get_next", MagicMock(return_value=run_time)):
            self.schedule.eval(now=run_time)

        ret = self.schedule.job_status("job1")
        self.assertEqual(ret["_error"], "Invalid cron string. Ignoring job job1.")

    def test_eval_when_invalid_date(self):
        """
        verify that scheduled job does not run
        and returns the right error
        """
        run_time = dateutil.parser.parse("11/29/2017 4:00pm")

        job = {
            "schedule": {"job1": {"function": "test.ping", "when": "13/29/2017 1:00pm"}}
        }

        # Add the job to the scheduler
        self.schedule.opts.update(job)

        # Evaluate 1 second before the run time
        self.schedule.eval(now=run_time)
        ret = self.schedule.job_status("job1")
        self.assertEqual(
            ret["_error"], "Invalid date string 13/29/2017 1:00pm. Ignoring job job1."
        )

    def test_eval_whens_grain_not_dict(self):
        """
        verify that scheduled job does not run
        and returns the right error
        """
        run_time = dateutil.parser.parse("11/29/2017 4:00pm")

        job = {"schedule": {"job1": {"function": "test.ping", "when": "tea time"}}}

        self.schedule.opts["grains"]["whens"] = ["tea time"]

        # Add the job to the scheduler
        self.schedule.opts.update(job)

        # Evaluate 1 second before the run time
        self.schedule.eval(now=run_time)
        ret = self.schedule.job_status("job1")
        self.assertEqual(
            ret["_error"], 'Grain "whens" must be a dict. Ignoring job job1.'
        )

    def test_eval_once_invalid_datestring(self):
        """
        verify that scheduled job does not run
        and returns the right error
        """
        job = {
            "schedule": {
                "job1": {"function": "test.ping", "once": "2017-13-13T13:00:00"}
            }
        }
        run_time = dateutil.parser.parse("12/13/2017 1:00pm")

        # Add the job to the scheduler
        self.schedule.opts.update(job)

        # Evaluate 1 second at the run time
        self.schedule.eval(now=run_time)
        ret = self.schedule.job_status("job1")
        _expected = (
            "Date string could not be parsed: "
            "2017-13-13T13:00:00, %Y-%m-%dT%H:%M:%S. "
            "Ignoring job job1."
        )
        self.assertEqual(ret["_error"], _expected)

    def test_eval_skip_during_range_invalid_date(self):
        """
        verify that scheduled job does not run
        and returns the right error
        """

        job = {
            "schedule": {
                "job1": {
                    "function": "test.ping",
                    "hours": 1,
                    "skip_during_range": {"start": "1:00pm", "end": "25:00pm"},
                }
            }
        }

        # Add the job to the scheduler
        self.schedule.opts.update(job)

        # eval at 3:00pm to prime, simulate minion start up.
        run_time = dateutil.parser.parse("11/29/2017 3:00pm")
        self.schedule.eval(now=run_time)
        ret = self.schedule.job_status("job1")

        # eval at 4:00pm to prime
        run_time = dateutil.parser.parse("11/29/2017 4:00pm")
        self.schedule.eval(now=run_time)
        ret = self.schedule.job_status("job1")
        _expected = (
            "Invalid date string for end in " "skip_during_range. Ignoring " "job job1."
        )
        self.assertEqual(ret["_error"], _expected)

    def test_eval_skip_during_range_end_before_start(self):
        """
        verify that scheduled job does not run
        and returns the right error
        """

        job = {
            "schedule": {
                "job1": {
                    "function": "test.ping",
                    "hours": 1,
                    "skip_during_range": {"start": "1:00pm", "end": "12:00pm"},
                }
            }
        }

        # Add the job to the scheduler
        self.schedule.opts.update(job)

        # eval at 3:00pm to prime, simulate minion start up.
        run_time = dateutil.parser.parse("11/29/2017 3:00pm")
        self.schedule.eval(now=run_time)
        ret = self.schedule.job_status("job1")

        # eval at 4:00pm to prime
        run_time = dateutil.parser.parse("11/29/2017 4:00pm")
        self.schedule.eval(now=run_time)
        ret = self.schedule.job_status("job1")
        _expected = (
            "schedule.handle_func: Invalid "
            "range, end must be larger than "
            "start. Ignoring job job1."
        )
        self.assertEqual(ret["_error"], _expected)

    def test_eval_skip_during_range_not_dict(self):
        """
        verify that scheduled job does not run
        and returns the right error
        """

        job = {
            "schedule": {
                "job1": {
                    "function": "test.ping",
                    "hours": 1,
                    "skip_during_range": ["start", "1:00pm", "end", "12:00pm"],
                }
            }
        }

        # Add the job to the scheduler
        self.schedule.opts.update(job)

        # eval at 3:00pm to prime, simulate minion start up.
        run_time = dateutil.parser.parse("11/29/2017 3:00pm")
        self.schedule.eval(now=run_time)
        ret = self.schedule.job_status("job1")

        # eval at 4:00pm to prime
        run_time = dateutil.parser.parse("11/29/2017 4:00pm")
        self.schedule.eval(now=run_time)
        ret = self.schedule.job_status("job1")
        _expected = (
            "schedule.handle_func: Invalid, "
            "range must be specified as a "
            "dictionary. Ignoring job job1."
        )
        self.assertEqual(ret["_error"], _expected)



# -*- coding: utf-8 -*-
"""
    :codeauthor: Pedro Algarvio (pedro@algarvio.me)


    tests.integration.modules.event
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

from __future__ import absolute_import

import threading
import time

import pytest
import salt.utils.event as event
from salt.ext.six.moves.queue import Empty, Queue
from tests.support.case import ModuleCase


@pytest.mark.windows_whitelisted
@pytest.mark.usefixtures("salt_sub_minion")
class EventModuleTest(ModuleCase):
    def __test_event_fire_master(self):
        events = Queue()

        def get_event(events):
            me = event.MasterEvent(self.master_opts["sock_dir"], listen=True)
            events.put_nowait(me.get_event(wait=10, tag="salttest", full=False))

        threading.Thread(target=get_event, args=(events,)).start()
        time.sleep(1)  # Allow multiprocessing.Process to start

        ret = self.run_function(
            "event.fire_master", ["event.fire_master: just test it!!!!", "salttest"]
        )
        self.assertTrue(ret)

        eventfired = events.get(block=True, timeout=10)
        self.assertIsNotNone(eventfired)
        self.assertIn("event.fire_master: just test it!!!!", eventfired["data"])

        ret = self.run_function(
            "event.fire_master",
            ["event.fire_master: just test it!!!!", "salttest-miss"],
        )
        self.assertTrue(ret)

        with self.assertRaises(Empty):
            eventfired = events.get(block=True, timeout=10)

    def __test_event_fire(self):
        events = Queue()

        def get_event(events):
            me = event.MinionEvent(self.minion_opts, listen=True)
            events.put_nowait(me.get_event(wait=10, tag="salttest", full=False))

        threading.Thread(target=get_event, args=(events,)).start()
        time.sleep(1)  # Allow multiprocessing.Process to start

        ret = self.run_function(
            "event.fire", ["event.fire: just test it!!!!", "salttest"]
        )
        self.assertTrue(ret)

        eventfired = events.get(block=True, timeout=10)
        self.assertIsNotNone(eventfired)
        self.assertIn("event.fire: just test it!!!!", eventfired)

        ret = self.run_function(
            "event.fire", ["event.fire: just test it!!!!", "salttest-miss"]
        )
        self.assertTrue(ret)

        with self.assertRaises(Empty):
            eventfired = events.get(block=True, timeout=10)

    def __test_event_fire_ipc_mode_tcp(self):
        events = Queue()

        def get_event(events):
            me = event.MinionEvent(self.sub_minion_opts, listen=True)
            events.put_nowait(me.get_event(wait=10, tag="salttest", full=False))

        threading.Thread(target=get_event, args=(events,)).start()
        time.sleep(1)  # Allow multiprocessing.Process to start

        ret = self.run_function(
            "event.fire",
            ["event.fire: just test it!!!!", "salttest"],
            minion_tgt="sub_minion",
        )
        self.assertTrue(ret)

        eventfired = events.get(block=True, timeout=10)
        self.assertIsNotNone(eventfired)
        self.assertIn("event.fire: just test it!!!!", eventfired)

        ret = self.run_function(
            "event.fire",
            ["event.fire: just test it!!!!", "salttest-miss"],
            minion_tgt="sub_minion",
        )
        self.assertTrue(ret)

        with self.assertRaises(Empty):
            eventfired = events.get(block=True, timeout=10)



# -*- coding: utf-8 -*-

# Import Python libs
from __future__ import absolute_import, print_function, unicode_literals

# Import Salt Libs
import salt.states.win_wusa as wusa
from salt.exceptions import SaltInvocationError

# Import Salt Testing Libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class WinWusaTestCase(TestCase, LoaderModuleMockMixin):
    """
    test the function in the win_wusa state module
    """

    kb = "KB123456"

    def setup_loader_modules(self):
        return {wusa: {"__opts__": {"test": False}, "__env__": "base"}}

    def test_installed_no_source(self):
        """
        test wusa.installed without passing source
        """
        with self.assertRaises(SaltInvocationError) as excinfo:
            wusa.installed(name="KB123456", source=None)

        self.assertEqual(
            excinfo.exception.strerror, 'Must specify a "source" file to install'
        )

    def test_installed_existing(self):
        """
        test wusa.installed when the kb is already installed
        """
        mock_installed = MagicMock(return_value=True)
        with patch.dict(wusa.__salt__, {"wusa.is_installed": mock_installed}):
            returned = wusa.installed(
                name=self.kb, source="salt://{0}.msu".format(self.kb)
            )
            expected = {
                "changes": {},
                "comment": "{0} already installed".format(self.kb),
                "name": self.kb,
                "result": True,
            }
            self.assertDictEqual(expected, returned)

    def test_installed_test_true(self):
        """
        test wusa.installed with test=True
        """
        mock_installed = MagicMock(return_value=False)
        with patch.dict(
            wusa.__salt__, {"wusa.is_installed": mock_installed}
        ), patch.dict(wusa.__opts__, {"test": True}):
            returned = wusa.installed(
                name=self.kb, source="salt://{0}.msu".format(self.kb)
            )
            expected = {
                "changes": {},
                "comment": "{0} would be installed".format(self.kb),
                "name": self.kb,
                "result": None,
            }
            self.assertDictEqual(expected, returned)

    def test_installed_cache_fail(self):
        """
        test wusa.install when it fails to cache the file
        """
        mock_installed = MagicMock(return_value=False)
        mock_cache = MagicMock(return_value="")
        with patch.dict(
            wusa.__salt__,
            {"wusa.is_installed": mock_installed, "cp.cache_file": mock_cache},
        ):
            returned = wusa.installed(
                name=self.kb, source="salt://{0}.msu".format(self.kb)
            )
            expected = {
                "changes": {},
                "comment": "Unable to cache salt://{0}.msu from "
                'saltenv "base"'.format(self.kb),
                "name": self.kb,
                "result": False,
            }
            self.assertDictEqual(expected, returned)

    def test_installed(self):
        """
        test wusa.installed assuming success
        """
        mock_installed = MagicMock(side_effect=[False, True])
        mock_cache = MagicMock(return_value="C:\\{0}.msu".format(self.kb))
        with patch.dict(
            wusa.__salt__,
            {
                "wusa.is_installed": mock_installed,
                "cp.cache_file": mock_cache,
                "wusa.install": MagicMock(),
            },
        ):
            returned = wusa.installed(
                name=self.kb, source="salt://{0}.msu".format(self.kb)
            )
            expected = {
                "changes": {"new": True, "old": False},
                "comment": "{0} was installed. ".format(self.kb),
                "name": self.kb,
                "result": True,
            }
            self.assertDictEqual(expected, returned)

    def test_installed_failed(self):
        """
        test wusa.installed with a failure
        """
        mock_installed = MagicMock(side_effect=[False, False])
        mock_cache = MagicMock(return_value="C:\\{0}.msu".format(self.kb))
        with patch.dict(
            wusa.__salt__,
            {
                "wusa.is_installed": mock_installed,
                "cp.cache_file": mock_cache,
                "wusa.install": MagicMock(),
            },
        ):
            returned = wusa.installed(
                name=self.kb, source="salt://{0}.msu".format(self.kb)
            )
            expected = {
                "changes": {},
                "comment": "{0} failed to install. ".format(self.kb),
                "name": self.kb,
                "result": False,
            }
            self.assertDictEqual(expected, returned)

    def test_uninstalled_non_existing(self):
        """
        test wusa.uninstalled when the kb is not installed
        """
        mock_installed = MagicMock(return_value=False)
        with patch.dict(wusa.__salt__, {"wusa.is_installed": mock_installed}):
            returned = wusa.uninstalled(name=self.kb)
            expected = {
                "changes": {},
                "comment": "{0} already uninstalled".format(self.kb),
                "name": self.kb,
                "result": True,
            }
            self.assertDictEqual(expected, returned)

    def test_uninstalled_test_true(self):
        """
        test wusa.uninstalled with test=True
        """
        mock_installed = MagicMock(return_value=True)
        with patch.dict(
            wusa.__salt__, {"wusa.is_installed": mock_installed}
        ), patch.dict(wusa.__opts__, {"test": True}):
            returned = wusa.uninstalled(name=self.kb)
            expected = {
                "changes": {},
                "comment": "{0} would be uninstalled".format(self.kb),
                "name": self.kb,
                "result": None,
            }
            self.assertDictEqual(expected, returned)

    def test_uninstalled(self):
        """
        test wusa.uninstalled assuming success
        """
        mock_installed = MagicMock(side_effect=[True, False])
        with patch.dict(
            wusa.__salt__,
            {"wusa.is_installed": mock_installed, "wusa.uninstall": MagicMock()},
        ):
            returned = wusa.uninstalled(name=self.kb)
            expected = {
                "changes": {"new": False, "old": True},
                "comment": "{0} was uninstalled".format(self.kb),
                "name": self.kb,
                "result": True,
            }
            self.assertDictEqual(expected, returned)

    def test_uninstalled_failed(self):
        """
        test wusa.uninstalled with a failure
        """
        mock_installed = MagicMock(side_effect=[True, True])
        with patch.dict(
            wusa.__salt__,
            {"wusa.is_installed": mock_installed, "wusa.uninstall": MagicMock()},
        ):
            returned = wusa.uninstalled(name=self.kb)
            expected = {
                "changes": {},
                "comment": "{0} failed to uninstall".format(self.kb),
                "name": self.kb,
                "result": False,
            }
            self.assertDictEqual(expected, returned)



# -*- coding: utf-8 -*-
"""
    :codeauthor: Thayne Harbaugh (tharbaug@adobe.com)

    tests.integration.shell.saltcli
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    :NOTE: this was named ``saltcli`` rather than ``salt`` because ``salt`` conflates
           in the python importer with the expected ``salt`` namespace and breaks imports.
"""

from __future__ import absolute_import

import logging
import os

import pytest
import salt.defaults.exitcodes
import salt.utils.files
import salt.utils.path
from tests.integration.utils import testprogram
from tests.support.case import ShellCase
from tests.support.helpers import slowTest
from tests.support.runtests import RUNTIME_VARS

log = logging.getLogger(__name__)


@pytest.mark.windows_whitelisted
class SaltTest(testprogram.TestProgramCase):
    """
    Various integration tests for the salt executable.
    """

    # pylint: disable=invalid-name
    @slowTest
    def test_exit_status_unknown_argument(self):
        """
        Ensure correct exit status when an unknown argument is passed to salt-run.
        """

        runner = testprogram.TestProgramSalt(
            name="run-unknown_argument", parent_dir=self._test_dir,
        )
        # Call setup here to ensure config and script exist
        runner.setup()
        stdout, stderr, status = runner.run(
            args=["--unknown-argument"], catch_stderr=True, with_retcode=True,
        )
        self.assert_exit_status(
            status, "EX_USAGE", message="unknown argument", stdout=stdout, stderr=stderr
        )
        # runner.shutdown() should be unnecessary since the start-up should fail

    @slowTest
    def test_exit_status_correct_usage(self):
        """
        Ensure correct exit status when salt-run starts correctly.
        """

        runner = testprogram.TestProgramSalt(
            name="run-correct_usage", parent_dir=self._test_dir,
        )
        # Call setup here to ensure config and script exist
        runner.setup()
        stdout, stderr, status = runner.run(
            args=["*", "-h"], catch_stderr=True, with_retcode=True,
        )
        self.assert_exit_status(
            status, "EX_OK", message="correct usage", stdout=stdout, stderr=stderr
        )


@pytest.mark.windows_whitelisted
class RetcodeTestCase(ShellCase):
    """
    Tests to ensure that we set non-zero retcodes when execution fails
    """

    # Hard-coding these instead of substituting values from
    # salt.defaults.exitcodes will give us a heads-up in the event that someone
    # tries to do something daft like change these values.
    error_status = salt.defaults.exitcodes.EX_GENERIC
    state_compiler_error = salt.defaults.exitcodes.EX_STATE_COMPILER_ERROR
    state_failure = salt.defaults.exitcodes.EX_STATE_FAILURE

    def _salt(self, command):
        cmdline = "minion " + command
        return self.run_salt(cmdline, with_retcode=True)[1]

    def _salt_call(self, command, retcode_passthrough=False):
        cmdline = "--retcode-passthrough " if retcode_passthrough else ""
        cmdline += command
        return self.run_call(cmdline, with_retcode=True)[1]

    def _test_error(self, salt_call=False):
        """
        Tests retcode when various error conditions are triggered
        """
        _run = self._salt_call if salt_call else self._salt

        retcode = _run("test.raise_exception TypeError")
        assert retcode == self.error_status, retcode

        retcode = _run("test.raise_exception salt.exceptions.CommandNotFoundError")
        assert retcode == self.error_status, retcode

        retcode = _run("test.raise_exception salt.exceptions.CommandExecutionError")
        assert retcode == self.error_status, retcode

        retcode = _run("test.raise_exception salt.exceptions.SaltInvocationError")
        assert retcode == self.error_status, retcode

        retcode = _run(
            "test.raise_exception " 'OSError 2 "No such file or directory" /tmp/foo.txt'
        )
        assert retcode == self.error_status, retcode

        retcode = _run('test.echo "{foo: bar, result: False}"')
        assert retcode == self.error_status, retcode

        retcode = _run('test.echo "{foo: bar, success: False}"')
        assert retcode == self.error_status, retcode

    @slowTest
    def test_zero_exit_code(self):
        """
        Test that a zero exit code is set when there are no errors and there is
        no explicit False result set in the return data.
        """
        retcode = self._salt("test.ping")
        assert retcode == 0, retcode

        retcode = self._salt_call("test.ping")
        assert retcode == 0, retcode

    @slowTest
    def test_context_retcode(self):
        """
        Test that a nonzero retcode set in the context dunder will cause the
        salt CLI to set a nonzero retcode.
        """
        # test.retcode will set the retcode in the context dunder
        retcode = self._salt("test.retcode 0")
        assert retcode == 0, retcode
        retcode = self._salt("test.retcode 42")
        assert retcode == self.error_status, retcode

        # Test salt-call, making sure to also confirm the behavior of
        # retcode_passthrough.
        retcode = self._salt_call("test.retcode 0")
        assert retcode == 0, retcode
        retcode = self._salt_call("test.retcode 42")
        assert retcode == self.error_status, retcode
        retcode = self._salt_call("test.retcode 42", retcode_passthrough=True)
        assert retcode == 42, retcode

        # Test a state run that exits with one or more failures
        retcode = self._salt_call("state.single test.fail_without_changes foo")
        assert retcode == self.error_status, retcode
        retcode = self._salt_call(
            "state.single test.fail_without_changes foo", retcode_passthrough=True
        )
        assert retcode == self.state_failure, retcode

        # Test a state compiler error
        retcode = self._salt_call("state.apply thisslsfiledoesnotexist")
        assert retcode == self.error_status, retcode
        retcode = self._salt_call(
            "state.apply thisslsfiledoesnotexist", retcode_passthrough=True
        )
        assert retcode == self.state_compiler_error, retcode

    @slowTest
    def test_salt_error(self):
        """
        Test that we return the expected retcode when a minion function raises
        an exception.
        """
        self._test_error()
        self._test_error(salt_call=True)

    @slowTest
    def test_missing_minion(self):
        """
        Test that a minion which doesn't respond results in a nonzeo exit code
        """
        good = salt.utils.path.join(self.master_opts["pki_dir"], "minions", "minion")
        bad = salt.utils.path.join(self.master_opts["pki_dir"], "minions", "minion2")
        try:
            # Copy the key
            with salt.utils.files.fopen(good, "rb") as fhr, salt.utils.files.fopen(
                bad, "wb"
            ) as fhw:
                fhw.write(fhr.read())
            retcode = self.run_script(
                "salt",
                "-c {0} -t 5 minion2 test.ping".format(RUNTIME_VARS.TMP_CONF_DIR),
                with_retcode=True,
                timeout=60,
            )[1]
            assert retcode == salt.defaults.exitcodes.EX_GENERIC, retcode
        finally:
            # Now get rid of it
            try:
                os.remove(bad)
            except OSError as exc:
                if exc.errno != os.errno.ENOENT:
                    log.error(
                        "Failed to remove %s, this may affect other tests: %s", bad, exc
                    )



# -*- coding: utf-8 -*-
"""
    :codeauthor: Nicole Thomas <nicole@saltstack.com>
"""
from __future__ import absolute_import, print_function, unicode_literals

import os
import shutil
import tempfile

import salt.exceptions
import salt.state
import salt.utils.files
import salt.utils.platform
from salt.exceptions import CommandExecutionError
from salt.utils.decorators import state as statedecorators
from salt.utils.odict import OrderedDict
from tests.support.helpers import slowTest, with_tempfile
from tests.support.mixins import AdaptedConfigurationTestCaseMixin
from tests.support.mock import MagicMock, patch
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import TestCase, skipIf

try:
    import pytest
except ImportError as err:
    pytest = None


class StateCompilerTestCase(TestCase, AdaptedConfigurationTestCaseMixin):
    """
    TestCase for the state compiler.
    """

    def test_format_log_non_ascii_character(self):
        """
        Tests running a non-ascii character through the state.format_log
        function. See Issue #33605.
        """
        # There is no return to test against as the format_log
        # function doesn't return anything. However, we do want
        # to make sure that the function doesn't stacktrace when
        # called.
        ret = {
            "changes": {"Français": {"old": "something old", "new": "something new"}},
            "result": True,
        }
        salt.state.format_log(ret)

    @slowTest
    def test_render_error_on_invalid_requisite(self):
        """
        Test that the state compiler correctly deliver a rendering
        exception when a requisite cannot be resolved
        """
        with patch("salt.state.State._gather_pillar") as state_patch:
            high_data = {
                "git": OrderedDict(
                    [
                        (
                            "pkg",
                            [
                                OrderedDict(
                                    [
                                        (
                                            "require",
                                            [
                                                OrderedDict(
                                                    [
                                                        (
                                                            "file",
                                                            OrderedDict(
                                                                [("test1", "test")]
                                                            ),
                                                        )
                                                    ]
                                                )
                                            ],
                                        )
                                    ]
                                ),
                                "installed",
                                {"order": 10000},
                            ],
                        ),
                        ("__sls__", "issue_35226"),
                        ("__env__", "base"),
                    ]
                )
            }
            minion_opts = self.get_temp_config("minion")
            minion_opts["pillar"] = {"git": OrderedDict([("test1", "test")])}
            state_obj = salt.state.State(minion_opts)
            with self.assertRaises(salt.exceptions.SaltRenderError):
                state_obj.call_high(high_data)

    def test_verify_onlyif_parse(self):
        low_data = {
            "onlyif": [{"fun": "test.arg", "args": ["arg1", "arg2"]}],
            "name": "mysql-server-5.7",
            "state": "debconf",
            "__id__": "set root password",
            "fun": "set",
            "__env__": "base",
            "__sls__": "debconf",
            "data": {
                "mysql-server/root_password": {"type": "password", "value": "temp123"}
            },
            "order": 10000,
        }
        expected_result = {"comment": "onlyif condition is true", "result": False}

        with patch("salt.state.State._gather_pillar") as state_patch:
            minion_opts = self.get_temp_config("minion")
            state_obj = salt.state.State(minion_opts)
            return_result = state_obj._run_check_onlyif(low_data, "")
            self.assertEqual(expected_result, return_result)

    def test_verify_onlyif_cmd_error(self):
        """
        Simulates a failure in cmd.retcode from onlyif
        This could occur is runas is specified with a user that does not exist
        """
        low_data = {
            "onlyif": "somecommand",
            "runas" "doesntexist" "name": "echo something",
            "state": "cmd",
            "__id__": "this is just a test",
            "fun": "run",
            "__env__": "base",
            "__sls__": "sometest",
            "order": 10000,
        }
        expected_result = {
            "comment": "onlyif condition is false",
            "result": True,
            "skip_watch": True,
        }

        with patch("salt.state.State._gather_pillar") as state_patch:
            minion_opts = self.get_temp_config("minion")
            state_obj = salt.state.State(minion_opts)
            mock = MagicMock(side_effect=CommandExecutionError("Boom!"))
            with patch.dict(state_obj.functions, {"cmd.retcode": mock}):
                #  The mock handles the exception, but the runas dict is being passed as it would actually be
                return_result = state_obj._run_check_onlyif(
                    low_data, {"runas": "doesntexist"}
                )
                self.assertEqual(expected_result, return_result)

    def test_verify_unless_cmd_error(self):
        """
        Simulates a failure in cmd.retcode from unless
        This could occur is runas is specified with a user that does not exist
        """
        low_data = {
            "unless": "somecommand",
            "runas" "doesntexist" "name": "echo something",
            "state": "cmd",
            "__id__": "this is just a test",
            "fun": "run",
            "__env__": "base",
            "__sls__": "sometest",
            "order": 10000,
        }
        expected_result = {
            "comment": "unless condition is true",
            "result": True,
            "skip_watch": True,
        }

        with patch("salt.state.State._gather_pillar") as state_patch:
            minion_opts = self.get_temp_config("minion")
            state_obj = salt.state.State(minion_opts)
            mock = MagicMock(side_effect=CommandExecutionError("Boom!"))
            with patch.dict(state_obj.functions, {"cmd.retcode": mock}):
                #  The mock handles the exception, but the runas dict is being passed as it would actually be
                return_result = state_obj._run_check_unless(
                    low_data, {"runas": "doesntexist"}
                )
                self.assertEqual(expected_result, return_result)

    def test_verify_unless_parse(self):
        low_data = {
            "unless": [{"fun": "test.arg", "args": ["arg1", "arg2"]}],
            "name": "mysql-server-5.7",
            "state": "debconf",
            "__id__": "set root password",
            "fun": "set",
            "__env__": "base",
            "__sls__": "debconf",
            "data": {
                "mysql-server/root_password": {"type": "password", "value": "temp123"}
            },
            "order": 10000,
        }
        expected_result = {
            "comment": "unless condition is true",
            "result": True,
            "skip_watch": True,
        }

        with patch("salt.state.State._gather_pillar") as state_patch:
            minion_opts = self.get_temp_config("minion")
            state_obj = salt.state.State(minion_opts)
            return_result = state_obj._run_check_unless(low_data, "")
            self.assertEqual(expected_result, return_result)

    def test_verify_creates(self):
        low_data = {
            "state": "cmd",
            "name": 'echo "something"',
            "__sls__": "tests.creates",
            "__env__": "base",
            "__id__": "do_a_thing",
            "creates": "/tmp/thing",
            "order": 10000,
            "fun": "run",
        }

        with patch("salt.state.State._gather_pillar") as state_patch:
            minion_opts = self.get_temp_config("minion")
            state_obj = salt.state.State(minion_opts)
            with patch("os.path.exists") as path_mock:
                path_mock.return_value = True
                expected_result = {
                    "comment": "/tmp/thing exists",
                    "result": True,
                    "skip_watch": True,
                }
                return_result = state_obj._run_check_creates(low_data)
                self.assertEqual(expected_result, return_result)

                path_mock.return_value = False
                expected_result = {
                    "comment": "Creates files not found",
                    "result": False,
                }
                return_result = state_obj._run_check_creates(low_data)
                self.assertEqual(expected_result, return_result)

    def test_verify_creates_list(self):
        low_data = {
            "state": "cmd",
            "name": 'echo "something"',
            "__sls__": "tests.creates",
            "__env__": "base",
            "__id__": "do_a_thing",
            "creates": ["/tmp/thing", "/tmp/thing2"],
            "order": 10000,
            "fun": "run",
        }

        with patch("salt.state.State._gather_pillar") as state_patch:
            minion_opts = self.get_temp_config("minion")
            state_obj = salt.state.State(minion_opts)
            with patch("os.path.exists") as path_mock:
                path_mock.return_value = True
                expected_result = {
                    "comment": "All files in creates exist",
                    "result": True,
                    "skip_watch": True,
                }
                return_result = state_obj._run_check_creates(low_data)
                self.assertEqual(expected_result, return_result)

                path_mock.return_value = False
                expected_result = {
                    "comment": "Creates files not found",
                    "result": False,
                }
                return_result = state_obj._run_check_creates(low_data)
                self.assertEqual(expected_result, return_result)

    def _expand_win_path(self, path):
        """
        Expand C:/users/admini~1/appdata/local/temp/salt-tests-tmpdir/...
        into C:/users/adminitrator/appdata/local/temp/salt-tests-tmpdir/...
        to prevent file.search from expanding the "~" with os.path.expanduser
        """
        if salt.utils.platform.is_windows():
            import win32file

            return win32file.GetLongPathName(path).replace("\\", "/")
        else:
            return path

    @with_tempfile()
    def test_verify_onlyif_parse_slots(self, name):
        with salt.utils.files.fopen(name, "w") as fp:
            fp.write("file-contents")
        low_data = {
            "onlyif": [
                {
                    "fun": "file.search",
                    "args": [
                        "__slot__:salt:test.echo({})".format(
                            self._expand_win_path(name)
                        ),
                    ],
                    "pattern": "__slot__:salt:test.echo(file-contents)",
                }
            ],
            "name": "mysql-server-5.7",
            "state": "debconf",
            "__id__": "set root password",
            "fun": "set",
            "__env__": "base",
            "__sls__": "debconf",
            "data": {
                "mysql-server/root_password": {"type": "password", "value": "temp123"}
            },
            "order": 10000,
        }
        expected_result = {"comment": "onlyif condition is true", "result": False}
        with patch("salt.state.State._gather_pillar") as state_patch:
            minion_opts = self.get_temp_config("minion")
            state_obj = salt.state.State(minion_opts)
            return_result = state_obj._run_check_onlyif(low_data, "")
            self.assertEqual(expected_result, return_result)

    def test_verify_onlyif_list_cmd(self):
        low_data = {
            "state": "cmd",
            "name": 'echo "something"',
            "__sls__": "tests.cmd",
            "__env__": "base",
            "__id__": "check onlyif",
            "onlyif": ["/bin/true", "/bin/false"],
            "order": 10001,
            "fun": "run",
        }
        expected_result = {
            "comment": "onlyif condition is false",
            "result": True,
            "skip_watch": True,
        }
        with patch("salt.state.State._gather_pillar") as state_patch:
            minion_opts = self.get_temp_config("minion")
            state_obj = salt.state.State(minion_opts)
            return_result = state_obj._run_check_onlyif(low_data, {})
            self.assertEqual(expected_result, return_result)

    @with_tempfile()
    def test_verify_unless_parse_slots(self, name):
        with salt.utils.files.fopen(name, "w") as fp:
            fp.write("file-contents")
        low_data = {
            "unless": [
                {
                    "fun": "file.search",
                    "args": [
                        "__slot__:salt:test.echo({})".format(
                            self._expand_win_path(name)
                        ),
                    ],
                    "pattern": "__slot__:salt:test.echo(file-contents)",
                }
            ],
            "name": "mysql-server-5.7",
            "state": "debconf",
            "__id__": "set root password",
            "fun": "set",
            "__env__": "base",
            "__sls__": "debconf",
            "data": {
                "mysql-server/root_password": {"type": "password", "value": "temp123"}
            },
            "order": 10000,
        }
        expected_result = {
            "comment": "unless condition is true",
            "result": True,
            "skip_watch": True,
        }

        with patch("salt.state.State._gather_pillar") as state_patch:
            minion_opts = self.get_temp_config("minion")
            state_obj = salt.state.State(minion_opts)
            return_result = state_obj._run_check_unless(low_data, "")
            self.assertEqual(expected_result, return_result)

    def test_verify_retry_parsing(self):
        low_data = {
            "state": "file",
            "name": "/tmp/saltstack.README.rst",
            "__sls__": "demo.download",
            "__env__": "base",
            "__id__": "download sample data",
            "retry": {"attempts": 5, "interval": 5},
            "unless": ["test -f /tmp/saltstack.README.rst"],
            "source": [
                "https://raw.githubusercontent.com/saltstack/salt/develop/README.rst"
            ],
            "source_hash": "f2bc8c0aa2ae4f5bb5c2051686016b48",
            "order": 10000,
            "fun": "managed",
        }
        expected_result = {
            "__id__": "download sample data",
            "__run_num__": 0,
            "__sls__": "demo.download",
            "changes": {},
            "comment": "['unless condition is true']  The state would be retried every 5 "
            "seconds (with a splay of up to 0 seconds) a maximum of 5 times or "
            "until a result of True is returned",
            "name": "/tmp/saltstack.README.rst",
            "result": True,
            "skip_watch": True,
        }

        with patch("salt.state.State._gather_pillar") as state_patch:
            minion_opts = self.get_temp_config("minion")
            minion_opts["test"] = True
            minion_opts["file_client"] = "local"
            state_obj = salt.state.State(minion_opts)
            mock = {
                "result": True,
                "comment": ["unless condition is true"],
                "skip_watch": True,
            }
            with patch.object(state_obj, "_run_check", return_value=mock):
                self.assertDictContainsSubset(expected_result, state_obj.call(low_data))

    def test_render_requisite_require_disabled(self):
        """
        Test that the state compiler correctly deliver a rendering
        exception when a requisite cannot be resolved
        """
        with patch("salt.state.State._gather_pillar") as state_patch:
            high_data = {
                "step_one": OrderedDict(
                    [
                        (
                            "test",
                            [
                                OrderedDict(
                                    [("require", [OrderedDict([("test", "step_two")])])]
                                ),
                                "succeed_with_changes",
                                {"order": 10000},
                            ],
                        ),
                        ("__sls__", "test.disable_require"),
                        ("__env__", "base"),
                    ]
                ),
                "step_two": {
                    "test": ["succeed_with_changes", {"order": 10001}],
                    "__env__": "base",
                    "__sls__": "test.disable_require",
                },
            }

            minion_opts = self.get_temp_config("minion")
            minion_opts["disabled_requisites"] = ["require"]
            state_obj = salt.state.State(minion_opts)
            ret = state_obj.call_high(high_data)
            run_num = ret["test_|-step_one_|-step_one_|-succeed_with_changes"][
                "__run_num__"
            ]
            self.assertEqual(run_num, 0)

    def test_render_requisite_require_in_disabled(self):
        """
        Test that the state compiler correctly deliver a rendering
        exception when a requisite cannot be resolved
        """
        with patch("salt.state.State._gather_pillar") as state_patch:
            high_data = {
                "step_one": {
                    "test": ["succeed_with_changes", {"order": 10000}],
                    "__env__": "base",
                    "__sls__": "test.disable_require_in",
                },
                "step_two": OrderedDict(
                    [
                        (
                            "test",
                            [
                                OrderedDict(
                                    [
                                        (
                                            "require_in",
                                            [OrderedDict([("test", "step_one")])],
                                        )
                                    ]
                                ),
                                "succeed_with_changes",
                                {"order": 10001},
                            ],
                        ),
                        ("__sls__", "test.disable_require_in"),
                        ("__env__", "base"),
                    ]
                ),
            }

            minion_opts = self.get_temp_config("minion")
            minion_opts["disabled_requisites"] = ["require_in"]
            state_obj = salt.state.State(minion_opts)
            ret = state_obj.call_high(high_data)
            run_num = ret["test_|-step_one_|-step_one_|-succeed_with_changes"][
                "__run_num__"
            ]
            self.assertEqual(run_num, 0)


class HighStateTestCase(TestCase, AdaptedConfigurationTestCaseMixin):
    def setUp(self):
        root_dir = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)
        self.state_tree_dir = os.path.join(root_dir, "state_tree")
        cache_dir = os.path.join(root_dir, "cachedir")
        for dpath in (root_dir, self.state_tree_dir, cache_dir):
            if not os.path.isdir(dpath):
                os.makedirs(dpath)

        overrides = {}
        overrides["root_dir"] = root_dir
        overrides["state_events"] = False
        overrides["id"] = "match"
        overrides["file_client"] = "local"
        overrides["file_roots"] = dict(base=[self.state_tree_dir])
        overrides["cachedir"] = cache_dir
        overrides["test"] = False
        self.config = self.get_temp_config("minion", **overrides)
        self.addCleanup(delattr, self, "config")
        self.highstate = salt.state.HighState(self.config)
        self.addCleanup(delattr, self, "highstate")
        self.highstate.push_active()

    def tearDown(self):
        self.highstate.pop_active()

    def test_top_matches_with_list(self):
        top = {"env": {"match": ["state1", "state2"], "nomatch": ["state3"]}}
        matches = self.highstate.top_matches(top)
        self.assertEqual(matches, {"env": ["state1", "state2"]})

    def test_top_matches_with_string(self):
        top = {"env": {"match": "state1", "nomatch": "state2"}}
        matches = self.highstate.top_matches(top)
        self.assertEqual(matches, {"env": ["state1"]})

    def test_matches_whitelist(self):
        matches = {"env": ["state1", "state2", "state3"]}
        matches = self.highstate.matches_whitelist(matches, ["state2"])
        self.assertEqual(matches, {"env": ["state2"]})

    def test_matches_whitelist_with_string(self):
        matches = {"env": ["state1", "state2", "state3"]}
        matches = self.highstate.matches_whitelist(matches, "state2,state3")
        self.assertEqual(matches, {"env": ["state2", "state3"]})

    def test_show_state_usage(self):
        # monkey patch sub methods
        self.highstate.avail = {"base": ["state.a", "state.b", "state.c"]}

        def verify_tops(*args, **kwargs):
            return []

        def get_top(*args, **kwargs):
            return None

        def top_matches(*args, **kwargs):
            return {"base": ["state.a", "state.b"]}

        self.highstate.verify_tops = verify_tops
        self.highstate.get_top = get_top
        self.highstate.top_matches = top_matches

        # get compile_state_usage() result
        state_usage_dict = self.highstate.compile_state_usage()

        self.assertEqual(state_usage_dict["base"]["count_unused"], 1)
        self.assertEqual(state_usage_dict["base"]["count_used"], 2)
        self.assertEqual(state_usage_dict["base"]["count_all"], 3)
        self.assertEqual(state_usage_dict["base"]["used"], ["state.a", "state.b"])
        self.assertEqual(state_usage_dict["base"]["unused"], ["state.c"])

    def test_find_sls_ids_with_exclude(self):
        """
        See https://github.com/saltstack/salt/issues/47182
        """
        sls_dir = "issue-47182"
        shutil.copytree(
            os.path.join(RUNTIME_VARS.BASE_FILES, sls_dir),
            os.path.join(self.state_tree_dir, sls_dir),
        )
        shutil.move(
            os.path.join(self.state_tree_dir, sls_dir, "top.sls"), self.state_tree_dir
        )
        # Manually compile the high data. We don't have to worry about all of
        # the normal error checking we do here since we know that all the SLS
        # files exist and there is no whitelist/blacklist being used.
        top = self.highstate.get_top()  # pylint: disable=assignment-from-none
        matches = self.highstate.top_matches(top)
        high, _ = self.highstate.render_highstate(matches)
        ret = salt.state.find_sls_ids("issue-47182.stateA.newer", high)
        self.assertEqual(ret, [("somestuff", "cmd")])


class MultiEnvHighStateTestCase(TestCase, AdaptedConfigurationTestCaseMixin):
    def setUp(self):
        root_dir = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)
        self.base_state_tree_dir = os.path.join(root_dir, "base")
        self.other_state_tree_dir = os.path.join(root_dir, "other")
        cache_dir = os.path.join(root_dir, "cachedir")
        for dpath in (
            root_dir,
            self.base_state_tree_dir,
            self.other_state_tree_dir,
            cache_dir,
        ):
            if not os.path.isdir(dpath):
                os.makedirs(dpath)
        shutil.copy(
            os.path.join(RUNTIME_VARS.BASE_FILES, "top.sls"), self.base_state_tree_dir
        )
        shutil.copy(
            os.path.join(RUNTIME_VARS.BASE_FILES, "core.sls"), self.base_state_tree_dir
        )
        shutil.copy(
            os.path.join(RUNTIME_VARS.BASE_FILES, "test.sls"), self.other_state_tree_dir
        )
        overrides = {}
        overrides["root_dir"] = root_dir
        overrides["state_events"] = False
        overrides["id"] = "match"
        overrides["file_client"] = "local"
        overrides["file_roots"] = dict(
            base=[self.base_state_tree_dir], other=[self.other_state_tree_dir]
        )
        overrides["cachedir"] = cache_dir
        overrides["test"] = False
        self.config = self.get_temp_config("minion", **overrides)
        self.addCleanup(delattr, self, "config")
        self.highstate = salt.state.HighState(self.config)
        self.addCleanup(delattr, self, "highstate")
        self.highstate.push_active()

    def tearDown(self):
        self.highstate.pop_active()

    def test_lazy_avail_states_base(self):
        # list_states not called yet
        self.assertEqual(self.highstate.avail._filled, False)
        self.assertEqual(self.highstate.avail._avail, {"base": None})
        # After getting 'base' env available states
        self.highstate.avail["base"]  # pylint: disable=pointless-statement
        self.assertEqual(self.highstate.avail._filled, False)
        self.assertEqual(self.highstate.avail._avail, {"base": ["core", "top"]})

    def test_lazy_avail_states_other(self):
        # list_states not called yet
        self.assertEqual(self.highstate.avail._filled, False)
        self.assertEqual(self.highstate.avail._avail, {"base": None})
        # After getting 'other' env available states
        self.highstate.avail["other"]  # pylint: disable=pointless-statement
        self.assertEqual(self.highstate.avail._filled, True)
        self.assertEqual(self.highstate.avail._avail, {"base": None, "other": ["test"]})

    def test_lazy_avail_states_multi(self):
        # list_states not called yet
        self.assertEqual(self.highstate.avail._filled, False)
        self.assertEqual(self.highstate.avail._avail, {"base": None})
        # After getting 'base' env available states
        self.highstate.avail["base"]  # pylint: disable=pointless-statement
        self.assertEqual(self.highstate.avail._filled, False)
        self.assertEqual(self.highstate.avail._avail, {"base": ["core", "top"]})
        # After getting 'other' env available states
        self.highstate.avail["other"]  # pylint: disable=pointless-statement
        self.assertEqual(self.highstate.avail._filled, True)
        self.assertEqual(
            self.highstate.avail._avail, {"base": ["core", "top"], "other": ["test"]}
        )


@skipIf(pytest is None, "PyTest is missing")
class StateReturnsTestCase(TestCase):
    """
    TestCase for code handling state returns.
    """

    def test_state_output_check_changes_is_dict(self):
        """
        Test that changes key contains a dictionary.
        :return:
        """
        data = {"changes": []}
        out = statedecorators.OutputUnifier("content_check")(lambda: data)()
        assert "'Changes' should be a dictionary" in out["comment"]
        assert not out["result"]

    def test_state_output_check_return_is_dict(self):
        """
        Test for the entire return is a dictionary
        :return:
        """
        data = ["whatever"]
        out = statedecorators.OutputUnifier("content_check")(lambda: data)()
        assert (
            "Malformed state return. Data must be a dictionary type" in out["comment"]
        )
        assert not out["result"]

    def test_state_output_check_return_has_nrc(self):
        """
        Test for name/result/comment keys are inside the return.
        :return:
        """
        data = {"arbitrary": "data", "changes": {}}
        out = statedecorators.OutputUnifier("content_check")(lambda: data)()
        assert (
            " The following keys were not present in the state return: name, result, comment"
            in out["comment"]
        )
        assert not out["result"]

    def test_state_output_unifier_comment_is_not_list(self):
        """
        Test for output is unified so the comment is converted to a multi-line string
        :return:
        """
        data = {
            "comment": ["data", "in", "the", "list"],
            "changes": {},
            "name": None,
            "result": "fantastic!",
        }
        expected = {
            "comment": "data\nin\nthe\nlist",
            "changes": {},
            "name": None,
            "result": True,
        }
        assert statedecorators.OutputUnifier("unify")(lambda: data)() == expected

        data = {
            "comment": ["data", "in", "the", "list"],
            "changes": {},
            "name": None,
            "result": None,
        }
        expected = "data\nin\nthe\nlist"
        assert (
            statedecorators.OutputUnifier("unify")(lambda: data)()["comment"]
            == expected
        )

    def test_state_output_unifier_result_converted_to_true(self):
        """
        Test for output is unified so the result is converted to True
        :return:
        """
        data = {
            "comment": ["data", "in", "the", "list"],
            "changes": {},
            "name": None,
            "result": "Fantastic",
        }
        assert statedecorators.OutputUnifier("unify")(lambda: data)()["result"] is True

    def test_state_output_unifier_result_converted_to_false(self):
        """
        Test for output is unified so the result is converted to False
        :return:
        """
        data = {
            "comment": ["data", "in", "the", "list"],
            "changes": {},
            "name": None,
            "result": "",
        }
        assert statedecorators.OutputUnifier("unify")(lambda: data)()["result"] is False


class StateFormatSlotsTestCase(TestCase, AdaptedConfigurationTestCaseMixin):
    """
    TestCase for code handling slots
    """

    def setUp(self):
        with patch("salt.state.State._gather_pillar"):
            minion_opts = self.get_temp_config("minion")
            self.state_obj = salt.state.State(minion_opts)

    def test_format_slots_no_slots(self):
        """
        Test the format slots keeps data without slots untouched.
        """
        cdata = {"args": ["arg"], "kwargs": {"key": "val"}}
        self.state_obj.format_slots(cdata)
        self.assertEqual(cdata, {"args": ["arg"], "kwargs": {"key": "val"}})

    @slowTest
    def test_format_slots_arg(self):
        """
        Test the format slots is calling a slot specified in args with corresponding arguments.
        """
        cdata = {
            "args": ["__slot__:salt:mod.fun(fun_arg, fun_key=fun_val)"],
            "kwargs": {"key": "val"},
        }
        mock = MagicMock(return_value="fun_return")
        with patch.dict(self.state_obj.functions, {"mod.fun": mock}):
            self.state_obj.format_slots(cdata)
        mock.assert_called_once_with("fun_arg", fun_key="fun_val")
        self.assertEqual(cdata, {"args": ["fun_return"], "kwargs": {"key": "val"}})

    @slowTest
    def test_format_slots_dict_arg(self):
        """
        Test the format slots is calling a slot specified in dict arg.
        """
        cdata = {
            "args": [{"subarg": "__slot__:salt:mod.fun(fun_arg, fun_key=fun_val)"}],
            "kwargs": {"key": "val"},
        }
        mock = MagicMock(return_value="fun_return")
        with patch.dict(self.state_obj.functions, {"mod.fun": mock}):
            self.state_obj.format_slots(cdata)
        mock.assert_called_once_with("fun_arg", fun_key="fun_val")
        self.assertEqual(
            cdata, {"args": [{"subarg": "fun_return"}], "kwargs": {"key": "val"}}
        )

    @slowTest
    def test_format_slots_listdict_arg(self):
        """
        Test the format slots is calling a slot specified in list containing a dict.
        """
        cdata = {
            "args": [[{"subarg": "__slot__:salt:mod.fun(fun_arg, fun_key=fun_val)"}]],
            "kwargs": {"key": "val"},
        }
        mock = MagicMock(return_value="fun_return")
        with patch.dict(self.state_obj.functions, {"mod.fun": mock}):
            self.state_obj.format_slots(cdata)
        mock.assert_called_once_with("fun_arg", fun_key="fun_val")
        self.assertEqual(
            cdata, {"args": [[{"subarg": "fun_return"}]], "kwargs": {"key": "val"}}
        )

    @slowTest
    def test_format_slots_liststr_arg(self):
        """
        Test the format slots is calling a slot specified in list containing a dict.
        """
        cdata = {
            "args": [["__slot__:salt:mod.fun(fun_arg, fun_key=fun_val)"]],
            "kwargs": {"key": "val"},
        }
        mock = MagicMock(return_value="fun_return")
        with patch.dict(self.state_obj.functions, {"mod.fun": mock}):
            self.state_obj.format_slots(cdata)
        mock.assert_called_once_with("fun_arg", fun_key="fun_val")
        self.assertEqual(cdata, {"args": [["fun_return"]], "kwargs": {"key": "val"}})

    @slowTest
    def test_format_slots_kwarg(self):
        """
        Test the format slots is calling a slot specified in kwargs with corresponding arguments.
        """
        cdata = {
            "args": ["arg"],
            "kwargs": {"key": "__slot__:salt:mod.fun(fun_arg, fun_key=fun_val)"},
        }
        mock = MagicMock(return_value="fun_return")
        with patch.dict(self.state_obj.functions, {"mod.fun": mock}):
            self.state_obj.format_slots(cdata)
        mock.assert_called_once_with("fun_arg", fun_key="fun_val")
        self.assertEqual(cdata, {"args": ["arg"], "kwargs": {"key": "fun_return"}})

    @slowTest
    def test_format_slots_multi(self):
        """
        Test the format slots is calling all slots with corresponding arguments when multiple slots
        specified.
        """
        cdata = {
            "args": [
                "__slot__:salt:test_mod.fun_a(a_arg, a_key=a_kwarg)",
                "__slot__:salt:test_mod.fun_b(b_arg, b_key=b_kwarg)",
            ],
            "kwargs": {
                "kw_key_1": "__slot__:salt:test_mod.fun_c(c_arg, c_key=c_kwarg)",
                "kw_key_2": "__slot__:salt:test_mod.fun_d(d_arg, d_key=d_kwarg)",
            },
        }
        mock_a = MagicMock(return_value="fun_a_return")
        mock_b = MagicMock(return_value="fun_b_return")
        mock_c = MagicMock(return_value="fun_c_return")
        mock_d = MagicMock(return_value="fun_d_return")
        with patch.dict(
            self.state_obj.functions,
            {
                "test_mod.fun_a": mock_a,
                "test_mod.fun_b": mock_b,
                "test_mod.fun_c": mock_c,
                "test_mod.fun_d": mock_d,
            },
        ):
            self.state_obj.format_slots(cdata)
        mock_a.assert_called_once_with("a_arg", a_key="a_kwarg")
        mock_b.assert_called_once_with("b_arg", b_key="b_kwarg")
        mock_c.assert_called_once_with("c_arg", c_key="c_kwarg")
        mock_d.assert_called_once_with("d_arg", d_key="d_kwarg")
        self.assertEqual(
            cdata,
            {
                "args": ["fun_a_return", "fun_b_return"],
                "kwargs": {"kw_key_1": "fun_c_return", "kw_key_2": "fun_d_return"},
            },
        )

    @slowTest
    def test_format_slots_malformed(self):
        """
        Test the format slots keeps malformed slots untouched.
        """
        sls_data = {
            "args": [
                "__slot__:NOT_SUPPORTED:not.called()",
                "__slot__:salt:not.called(",
                "__slot__:salt:",
                "__slot__:salt",
                "__slot__:",
                "__slot__",
            ],
            "kwargs": {
                "key3": "__slot__:NOT_SUPPORTED:not.called()",
                "key4": "__slot__:salt:not.called(",
                "key5": "__slot__:salt:",
                "key6": "__slot__:salt",
                "key7": "__slot__:",
                "key8": "__slot__",
            },
        }
        cdata = sls_data.copy()
        mock = MagicMock(return_value="return")
        with patch.dict(self.state_obj.functions, {"not.called": mock}):
            self.state_obj.format_slots(cdata)
        mock.assert_not_called()
        self.assertEqual(cdata, sls_data)

    @slowTest
    def test_slot_traverse_dict(self):
        """
        Test the slot parsing of dict response.
        """
        cdata = {
            "args": ["arg"],
            "kwargs": {"key": "__slot__:salt:mod.fun(fun_arg, fun_key=fun_val).key1"},
        }
        return_data = {"key1": "value1"}
        mock = MagicMock(return_value=return_data)
        with patch.dict(self.state_obj.functions, {"mod.fun": mock}):
            self.state_obj.format_slots(cdata)
        mock.assert_called_once_with("fun_arg", fun_key="fun_val")
        self.assertEqual(cdata, {"args": ["arg"], "kwargs": {"key": "value1"}})

    @slowTest
    def test_slot_append(self):
        """
        Test the slot parsing of dict response.
        """
        cdata = {
            "args": ["arg"],
            "kwargs": {
                "key": "__slot__:salt:mod.fun(fun_arg, fun_key=fun_val).key1 ~ thing~",
            },
        }
        return_data = {"key1": "value1"}
        mock = MagicMock(return_value=return_data)
        with patch.dict(self.state_obj.functions, {"mod.fun": mock}):
            self.state_obj.format_slots(cdata)
        mock.assert_called_once_with("fun_arg", fun_key="fun_val")
        self.assertEqual(cdata, {"args": ["arg"], "kwargs": {"key": "value1thing~"}})

    # Skip on windows like integration.modules.test_state.StateModuleTest.test_parallel_state_with_long_tag
    @skipIf(
        salt.utils.platform.is_windows(),
        "Skipped until parallel states can be fixed on Windows",
    )
    def test_format_slots_parallel(self):
        """
        Test if slots work with "parallel: true".
        """
        high_data = {
            "always-changes-and-succeeds": {
                "test": [
                    {"changes": True},
                    {"comment": "__slot__:salt:test.echo(fun_return)"},
                    {"parallel": True},
                    "configurable_test_state",
                    {"order": 10000},
                ],
                "__env__": "base",
                "__sls__": "parallel_slots",
            }
        }
        self.state_obj.jid = "123"
        res = self.state_obj.call_high(high_data)
        self.state_obj.jid = None
        [(_, data)] = res.items()
        self.assertEqual(data["comment"], "fun_return")



# -*- coding: utf-8 -*-
"""
Platform independent versions of some os/os.path functions. Gets around PY2's
lack of support for reading NTFS links.
"""

# Import python libs
from __future__ import absolute_import, print_function, unicode_literals

import errno
import logging
import os
import posixpath
import re
import string
import struct
from collections.abc import Iterable

# Import Salt libs
import salt.utils.args
import salt.utils.platform
import salt.utils.stringutils
from salt.exceptions import CommandNotFoundError

# Import 3rd-party libs
from salt.ext import six
from salt.utils.decorators.jinja import jinja_filter

try:
    import win32file
    from pywintypes import error as pywinerror

    HAS_WIN32FILE = True
except ImportError:
    HAS_WIN32FILE = False

log = logging.getLogger(__name__)


def islink(path):
    """
    Equivalent to os.path.islink()
    """
    if six.PY3 or not salt.utils.platform.is_windows():
        return os.path.islink(path)

    if not HAS_WIN32FILE:
        log.error("Cannot check if %s is a link, missing required modules", path)

    if not _is_reparse_point(path):
        return False

    # check that it is a symlink reparse point (in case it is something else,
    # like a mount point)
    reparse_data = _get_reparse_data(path)

    # sanity check - this should not happen
    if not reparse_data:
        # not a reparse point
        return False

    # REPARSE_DATA_BUFFER structure - see
    # http://msdn.microsoft.com/en-us/library/ff552012.aspx

    # parse the structure header to work out which type of reparse point this is
    header_parser = struct.Struct("L")
    (ReparseTag,) = header_parser.unpack(reparse_data[: header_parser.size])
    # http://msdn.microsoft.com/en-us/library/windows/desktop/aa365511.aspx
    if not ReparseTag & 0xA000FFFF == 0xA000000C:
        return False
    else:
        return True


def readlink(path):
    """
    Equivalent to os.readlink()
    """
    if six.PY3 or not salt.utils.platform.is_windows():
        return os.readlink(path)

    if not HAS_WIN32FILE:
        log.error("Cannot read %s, missing required modules", path)

    reparse_data = _get_reparse_data(path)

    if not reparse_data:
        # Reproduce *NIX behavior when os.readlink is performed on a path that
        # is not a symbolic link.
        raise OSError(errno.EINVAL, "Invalid argument: '{0}'".format(path))

    # REPARSE_DATA_BUFFER structure - see
    # http://msdn.microsoft.com/en-us/library/ff552012.aspx

    # parse the structure header to work out which type of reparse point this is
    header_parser = struct.Struct("L")
    (ReparseTag,) = header_parser.unpack(reparse_data[: header_parser.size])
    # http://msdn.microsoft.com/en-us/library/windows/desktop/aa365511.aspx
    if not ReparseTag & 0xA000FFFF == 0xA000000C:
        raise OSError(
            errno.EINVAL,
            "{0} is not a symlink, but another type of reparse point "
            "(0x{0:X}).".format(ReparseTag),
        )

    # parse as a symlink reparse point structure (the structure for other
    # reparse points is different)
    data_parser = struct.Struct("LHHHHHHL")
    (
        ReparseTag,
        ReparseDataLength,
        Reserved,
        SubstituteNameOffset,
        SubstituteNameLength,
        PrintNameOffset,
        PrintNameLength,
        Flags,
    ) = data_parser.unpack(reparse_data[: data_parser.size])

    path_buffer_offset = data_parser.size
    absolute_substitute_name_offset = path_buffer_offset + SubstituteNameOffset
    target_bytes = reparse_data[
        absolute_substitute_name_offset : absolute_substitute_name_offset
        + SubstituteNameLength
    ]
    target = target_bytes.decode("UTF-16")

    if target.startswith("\\??\\"):
        target = target[4:]

    try:
        # comes out in 8.3 form; convert it to LFN to make it look nicer
        target = win32file.GetLongPathName(target)
    except pywinerror as exc:
        # If target is on a UNC share, the decoded target will be in the format
        # "UNC\hostanme\sharename\additional\subdirs\under\share". So, in
        # these cases, return the target path in the proper UNC path format.
        if target.startswith("UNC\\"):
            return re.sub(r"^UNC\\+", r"\\\\", target)
        # if file is not found (i.e. bad symlink), return it anyway like on *nix
        if exc.winerror == 2:
            return target
        raise

    return target


def _is_reparse_point(path):
    """
    Returns True if path is a reparse point; False otherwise.
    """
    result = win32file.GetFileAttributesW(path)

    if result == -1:
        return False

    return True if result & 0x400 else False


def _get_reparse_data(path):
    """
    Retrieves the reparse point data structure for the given path.

    If the path is not a reparse point, None is returned.

    See http://msdn.microsoft.com/en-us/library/ff552012.aspx for details on the
    REPARSE_DATA_BUFFER structure returned.
    """
    # ensure paths are using the right slashes
    path = os.path.normpath(path)

    if not _is_reparse_point(path):
        return None

    fileHandle = None
    try:
        fileHandle = win32file.CreateFileW(
            path,
            0x80000000,  # GENERIC_READ
            1,  # share with other readers
            None,  # no inherit, default security descriptor
            3,  # OPEN_EXISTING
            0x00200000
            | 0x02000000,  # FILE_FLAG_OPEN_REPARSE_POINT | FILE_FLAG_BACKUP_SEMANTICS
        )

        reparseData = win32file.DeviceIoControl(
            fileHandle,
            0x900A8,  # FSCTL_GET_REPARSE_POINT
            None,  # in buffer
            16384,  # out buffer size (MAXIMUM_REPARSE_DATA_BUFFER_SIZE)
        )

    finally:
        if fileHandle:
            win32file.CloseHandle(fileHandle)

    return reparseData


@jinja_filter("which")
def which(exe=None):
    """
    Python clone of /usr/bin/which
    """

    if not exe:
        log.error("No executable was passed to be searched by salt.utils.path.which()")
        return None

    ## define some utilities (we use closures here because our predecessor used them)
    def is_executable_common(path):
        """
        This returns truth if posixy semantics (which python simulates on
        windows) states that this is executable.
        """
        return os.path.isfile(path) and os.access(path, os.X_OK)

    def resolve(path):
        """
        This will take a path and recursively follow the link until we get to a
        real file.
        """
        while os.path.islink(path):
            res = os.readlink(path)

            # if the link points to a relative target, then convert it to an
            # absolute path relative to the original path
            if not os.path.isabs(res):
                directory, _ = os.path.split(path)
                res = join(directory, res)
            path = res
        return path

    # windows-only
    def has_executable_ext(path, ext_membership):
        """
        Extract the extension from the specified path, lowercase it so we
        can be insensitive, and then check it against the available exts.
        """
        p, ext = os.path.splitext(path)
        return ext.lower() in ext_membership

    ## prepare related variables from the environment
    res = salt.utils.stringutils.to_unicode(os.environ.get("PATH", ""))
    system_path = res.split(os.pathsep)

    # add some reasonable defaults in case someone's PATH is busted
    if not salt.utils.platform.is_windows():
        res = set(system_path)
        extended_path = [
            "/sbin",
            "/bin",
            "/usr/sbin",
            "/usr/bin",
            "/usr/local/sbin",
            "/usr/local/bin",
        ]
        system_path.extend([p for p in extended_path if p not in res])

    ## now to define the semantics of what's considered executable on a given platform
    if salt.utils.platform.is_windows():
        # executable semantics on windows requires us to search PATHEXT
        res = salt.utils.stringutils.to_str(os.environ.get("PATHEXT", str(".EXE")))

        # generate two variables, one of them for O(n) searches (but ordered)
        # and another for O(1) searches. the previous guy was trying to use
        # memoization with a function that has no arguments, this provides
        # the exact same benefit
        pathext = res.split(os.pathsep)
        res = {ext.lower() for ext in pathext}

        # check if our caller already specified a valid extension as then we don't need to match it
        _, ext = os.path.splitext(exe)
        if ext.lower() in res:
            pathext = [""]

            is_executable = is_executable_common

        # The specified extension isn't valid, so we just assume it's part of the
        # filename and proceed to walk the pathext list
        else:
            is_executable = lambda path, membership=res: is_executable_common(
                path
            ) and has_executable_ext(path, membership)

    else:
        # in posix, there's no such thing as file extensions..only zuul
        pathext = [""]

        # executable semantics are pretty simple on reasonable platforms...
        is_executable = is_executable_common

    ## search for the executable

    # check to see if the full path was specified as then we don't need
    # to actually walk the system_path for any reason
    if is_executable(exe):
        return exe

    # now to search through our system_path
    for path in system_path:
        p = join(path, exe)

        # iterate through all extensions to see which one is executable
        for ext in pathext:
            pext = p + ext
            rp = resolve(pext)
            if is_executable(rp):
                return p + ext
            continue
        continue

    ## if something was executable, we should've found it already...
    log.trace(
        "'%s' could not be found in the following search path: '%s'", exe, system_path
    )
    return None


def which_bin(exes):
    """
    Scan over some possible executables and return the first one that is found
    """
    if not isinstance(exes, Iterable):
        return None
    for exe in exes:
        path = which(exe)
        if not path:
            continue
        return path
    return None


@jinja_filter("path_join")
def join(*parts, **kwargs):
    """
    This functions tries to solve some issues when joining multiple absolute
    paths on both *nix and windows platforms.

    See tests/unit/utils/path_join_test.py for some examples on what's being
    talked about here.

    The "use_posixpath" kwarg can be be used to force joining using poxixpath,
    which is useful for Salt fileserver paths on Windows masters.
    """
    if six.PY3:
        new_parts = []
        for part in parts:
            new_parts.append(salt.utils.stringutils.to_str(part))
        parts = new_parts

    kwargs = salt.utils.args.clean_kwargs(**kwargs)
    use_posixpath = kwargs.pop("use_posixpath", False)
    if kwargs:
        salt.utils.args.invalid_kwargs(kwargs)

    pathlib = posixpath if use_posixpath else os.path

    # Normalize path converting any os.sep as needed
    parts = [pathlib.normpath(p) for p in parts]

    try:
        root = parts.pop(0)
    except IndexError:
        # No args passed to func
        return ""

    root = salt.utils.stringutils.to_unicode(root)
    if not parts:
        ret = root
    else:
        stripped = [p.lstrip(os.sep) for p in parts]
        ret = pathlib.join(root, *salt.utils.data.decode(stripped))
    return pathlib.normpath(ret)


def check_or_die(command):
    """
    Simple convenience function for modules to use for gracefully blowing up
    if a required tool is not available in the system path.

    Lazily import `salt.modules.cmdmod` to avoid any sort of circular
    dependencies.
    """
    if command is None:
        raise CommandNotFoundError("'None' is not a valid command.")

    if not which(command):
        raise CommandNotFoundError("'{0}' is not in the path".format(command))


def sanitize_win_path(winpath):
    """
    Remove illegal path characters for windows
    """
    intab = "<>:|?*"
    if isinstance(winpath, six.text_type):
        winpath = winpath.translate(dict((ord(c), "_") for c in intab))
    elif isinstance(winpath, six.string_types):
        outtab = "_" * len(intab)
        trantab = (
            "".maketrans(intab, outtab) if six.PY3 else string.maketrans(intab, outtab)
        )  # pylint: disable=no-member
        winpath = winpath.translate(trantab)
    return winpath


def safe_path(path, allow_path=None):
    r"""
    .. versionadded:: 2017.7.3

    Checks that the path is safe for modification by Salt. For example, you
    wouldn't want to have salt delete the contents of ``C:\Windows``. The
    following directories are considered unsafe:

    - C:\, D:\, E:\, etc.
    - \
    - C:\Windows

    Args:

        path (str): The path to check

        allow_paths (str, list): A directory or list of directories inside of
            path that may be safe. For example: ``C:\Windows\TEMP``

    Returns:
        bool: True if safe, otherwise False
    """
    # Create regex definitions for directories that may be unsafe to modify
    system_root = os.environ.get("SystemRoot", "C:\\Windows")
    deny_paths = (
        r"[a-z]\:\\$",  # C:\, D:\, etc
        r"\\$",  # \
        re.escape(system_root),  # C:\Windows
    )

    # Make allow_path a list
    if allow_path and not isinstance(allow_path, list):
        allow_path = [allow_path]

    # Create regex definition for directories we may want to make exceptions for
    allow_paths = list()
    if allow_path:
        for item in allow_path:
            allow_paths.append(re.escape(item))

    # Check the path to make sure it's not one of the bad paths
    good_path = True
    for d_path in deny_paths:
        if re.match(d_path, path, flags=re.IGNORECASE) is not None:
            # Found deny path
            good_path = False

    # If local_dest is one of the bad paths, check for exceptions
    if not good_path:
        for a_path in allow_paths:
            if re.match(a_path, path, flags=re.IGNORECASE) is not None:
                # Found exception
                good_path = True

    return good_path


def os_walk(top, *args, **kwargs):
    """
    This is a helper than ensures that all paths returned from os.walk are
    unicode.
    """
    if six.PY2 and salt.utils.platform.is_windows():
        top_query = top
    else:
        top_query = salt.utils.stringutils.to_str(top)
    for item in os.walk(top_query, *args, **kwargs):
        yield salt.utils.data.decode(item, preserve_tuples=True)



# -*- coding: utf-8 -*-
"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

# Import Python Libs
from __future__ import absolute_import

import os

# Import Salt Libs
import salt.modules.poudriere as poudriere

# Import Salt Testing Libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, mock_open, patch
from tests.support.unit import TestCase


class PoudriereTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.poudriere
    """

    def setup_loader_modules(self):
        return {poudriere: {}}

    # 'is_jail' function tests: 1

    def test_is_jail(self):
        """
        Test if it return True if jail exists False if not.
        """
        mock = MagicMock(return_value="salt stack")
        with patch.dict(poudriere.__salt__, {"cmd.run": mock}), patch(
            "salt.modules.poudriere._check_config_exists", MagicMock(return_value=True)
        ):
            self.assertTrue(poudriere.is_jail("salt"))

            self.assertFalse(poudriere.is_jail("SALT"))

    # 'make_pkgng_aware' function tests: 1

    def test_make_pkgng_aware(self):
        """
        Test if it make jail ``jname`` pkgng aware.
        """
        temp_dir = os.path.join("tmp", "salt")
        conf_file = os.path.join("tmp", "salt", "salt-make.conf")
        ret1 = "Could not create or find required directory {0}".format(temp_dir)
        ret2 = "Looks like file {0} could not be created".format(conf_file)
        ret3 = {"changes": "Created {0}".format(conf_file)}
        mock = MagicMock(return_value=temp_dir)
        mock_true = MagicMock(return_value=True)
        with patch.dict(
            poudriere.__salt__, {"config.option": mock, "file.write": mock_true}
        ):
            with patch.object(os.path, "isdir", MagicMock(return_value=False)):
                with patch.object(os, "makedirs", mock_true):
                    self.assertEqual(poudriere.make_pkgng_aware("salt"), ret1)

            with patch.object(os.path, "isdir", mock_true):
                self.assertEqual(poudriere.make_pkgng_aware("salt"), ret2)

                with patch.object(os.path, "isfile", mock_true):
                    self.assertDictEqual(poudriere.make_pkgng_aware("salt"), ret3)

    # 'parse_config' function tests: 1

    def test_parse_config(self):
        """
        Test if it returns a dict of poudriere main configuration definitions.
        """
        mock = MagicMock(return_value="/tmp/salt")
        with patch.dict(poudriere.__salt__, {"config.option": mock}), patch(
            "salt.utils.files.fopen", mock_open()
        ), patch.object(
            poudriere, "_check_config_exists", MagicMock(side_effect=[True, False])
        ):
            self.assertDictEqual(poudriere.parse_config(), {})

            self.assertEqual(
                poudriere.parse_config(), "Could not find /tmp/salt on file system"
            )

    # 'version' function tests: 1

    def test_version(self):
        """
        Test if it return poudriere version.
        """
        mock = MagicMock(return_value="9.0-RELEASE")
        with patch.dict(poudriere.__salt__, {"cmd.run": mock}):
            self.assertEqual(poudriere.version(), "9.0-RELEASE")

    # 'list_jails' function tests: 1

    def test_list_jails(self):
        """
        Test if it return a list of current jails managed by poudriere.
        """
        mock = MagicMock(return_value="salt stack")
        with patch.dict(poudriere.__salt__, {"cmd.run": mock}), patch(
            "salt.modules.poudriere._check_config_exists", MagicMock(return_value=True)
        ):
            self.assertListEqual(poudriere.list_jails(), ["salt stack"])

    # 'list_ports' function tests: 1

    def test_list_ports(self):
        """
        Test if it return a list of current port trees managed by poudriere.
        """
        mock = MagicMock(return_value="salt stack")
        with patch.dict(poudriere.__salt__, {"cmd.run": mock}), patch(
            "salt.modules.poudriere._check_config_exists", MagicMock(return_value=True)
        ):
            self.assertListEqual(poudriere.list_ports(), ["salt stack"])

    # 'create_jail' function tests: 1

    def test_create_jail(self):
        """
        Test if it creates a new poudriere jail if one does not exist.
        """
        mock_stack = MagicMock(return_value="90amd64 stack")
        mock_true = MagicMock(return_value=True)
        with patch.dict(poudriere.__salt__, {"cmd.run": mock_stack}), patch(
            "salt.modules.poudriere._check_config_exists", MagicMock(return_value=True)
        ):
            self.assertEqual(
                poudriere.create_jail("90amd64", "amd64"), "90amd64 already exists"
            )

            with patch.object(poudriere, "make_pkgng_aware", mock_true):
                self.assertEqual(
                    poudriere.create_jail("80amd64", "amd64"),
                    "Issue creating jail 80amd64",
                )

        with patch.object(poudriere, "make_pkgng_aware", mock_true), patch(
            "salt.modules.poudriere._check_config_exists", MagicMock(return_value=True)
        ):
            with patch.object(
                poudriere, "is_jail", MagicMock(side_effect=[False, True])
            ):
                with patch.dict(poudriere.__salt__, {"cmd.run": mock_stack}):
                    self.assertEqual(
                        poudriere.create_jail("80amd64", "amd64"),
                        "Created jail 80amd64",
                    )

    # 'update_jail' function tests: 1

    def test_update_jail(self):
        """
        Test if it run freebsd-update on `name` poudriere jail.
        """
        mock = MagicMock(return_value="90amd64 stack")
        with patch.dict(poudriere.__salt__, {"cmd.run": mock}), patch(
            "salt.modules.poudriere._check_config_exists", MagicMock(return_value=True)
        ):
            self.assertEqual(poudriere.update_jail("90amd64"), "90amd64 stack")

            self.assertEqual(
                poudriere.update_jail("80amd64"), "Could not find jail 80amd64"
            )

    # 'delete_jail' function tests: 1

    def test_delete_jail(self):
        """
        Test if it deletes poudriere jail with `name`.
        """
        ret = "Looks like there was an issue deleteing jail             90amd64"
        mock_stack = MagicMock(return_value="90amd64 stack")
        with patch.dict(poudriere.__salt__, {"cmd.run": mock_stack}), patch(
            "salt.modules.poudriere._check_config_exists", MagicMock(return_value=True)
        ):
            self.assertEqual(poudriere.delete_jail("90amd64"), ret)

            self.assertEqual(
                poudriere.delete_jail("80amd64"),
                "Looks like jail 80amd64 has not been created",
            )

        ret1 = 'Deleted jail "80amd64" but was unable to remove jail make file'
        with patch.object(
            poudriere, "is_jail", MagicMock(side_effect=[True, False, True, False])
        ):
            with patch.dict(poudriere.__salt__, {"cmd.run": mock_stack}):
                with patch.object(
                    poudriere, "_config_dir", MagicMock(return_value="/tmp/salt")
                ):
                    self.assertEqual(
                        poudriere.delete_jail("80amd64"), "Deleted jail 80amd64"
                    )

                    with patch.object(os.path, "isfile", MagicMock(return_value=True)):
                        self.assertEqual(poudriere.delete_jail("80amd64"), ret1)

    # 'create_ports_tree' function tests: 1

    def test_create_ports_tree(self):
        """
        Test if it not working need to run portfetch non interactive.
        """
        mock = MagicMock(return_value="salt stack")
        with patch.dict(poudriere.__salt__, {"cmd.run": mock}), patch(
            "salt.modules.poudriere._check_config_exists", MagicMock(return_value=True)
        ):
            self.assertEqual(poudriere.create_ports_tree(), "salt stack")

    # 'update_ports_tree' function tests: 1

    def test_update_ports_tree(self):
        """
        Test if it updates the ports tree, either the default
        or the `ports_tree` specified.
        """
        mock = MagicMock(return_value="salt stack")
        with patch.dict(poudriere.__salt__, {"cmd.run": mock}), patch(
            "salt.modules.poudriere._check_config_exists", MagicMock(return_value=True)
        ):
            self.assertEqual(poudriere.update_ports_tree("staging"), "salt stack")

    # 'bulk_build' function tests: 1

    def test_bulk_build(self):
        """
        Test if it run bulk build on poudriere server.
        """
        ret = "Could not find file /root/pkg_list on filesystem"
        mock = MagicMock(return_value="salt stack")
        with patch.dict(poudriere.__salt__, {"cmd.run": mock}), patch(
            "salt.modules.poudriere._check_config_exists", MagicMock(return_value=True)
        ):
            self.assertEqual(poudriere.bulk_build("90amd64", "/root/pkg_list"), ret)

            with patch.object(os.path, "isfile", MagicMock(return_value=True)):
                self.assertEqual(
                    poudriere.bulk_build("90amd64", "/root/pkg_list"),
                    "Could not find jail 90amd64",
                )

        ret = (
            "There may have been an issue building "
            "packages dumping output: 90amd64 stack"
        )
        with patch.object(os.path, "isfile", MagicMock(return_value=True)), patch(
            "salt.modules.poudriere._check_config_exists", MagicMock(return_value=True)
        ):
            mock = MagicMock(return_value="90amd64 stack packages built")
            with patch.dict(poudriere.__salt__, {"cmd.run": mock}):
                self.assertEqual(
                    poudriere.bulk_build("90amd64", "/root/pkg_list"),
                    "90amd64 stack packages built",
                )

            mock = MagicMock(return_value="90amd64 stack")
            with patch.dict(poudriere.__salt__, {"cmd.run": mock}):
                self.assertEqual(poudriere.bulk_build("90amd64", "/root/pkg_list"), ret)
